import copy
import unittest

import pendulum
from azul_bedrock import models_restapi

from azul_client.api import Api
from azul_client.api.binaries_meta import FindOptions
from azul_client.config import get_config

from . import module_ref, module_sha256s, module_source


class TestBinaryFind(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.non_existent_sha256 = "ed7002b439e9ac845f22357d822bac1444730fbdb6016d3ec9432297b9ec9f74"
        cls.api = Api(get_config())
        cls.source = module_source
        cls.ref = module_ref
        cls.sha256s = copy.copy(module_sha256s)
        cls.default_min_hash_count = 10
        cls.source1 = "virustotal"
        cls.source2 = "watch"
        cls.common_plugin_name = "Entropy"
        cls.common_feature_value = "txt"
        cls.common_feature_key = "malformed"
        cls.common_file_format = "text/plain"
        cls.dummy_tag = "dummy"
        cls.user = module_ref.get("user")

    # -------------------------------- Hash Query

    def test_find_hashes(self):
        result = self.api.binaries_meta.find_hashes([self.sha256s[0]])
        self.assertEqual(1, len(result.items))
        self.assertEqual(self.sha256s[0], result.items[0].sha256)
        self.assertTrue(result.items[0].exists)

        sub_sha256s = self.sha256s[0:5]
        result = self.api.binaries_meta.find_hashes(sub_sha256s)
        self.assertEqual(5, len(result.items))
        # Verify that all of the sha256's found are equal to the ones in the list.
        found_hashes = set()
        for item in result.items:
            self.assertIn(item.sha256, sub_sha256s)
            found_hashes.add(item.sha256)
        # Ensure all found hashes are unique.
        self.assertEqual(5, len(found_hashes))

        result = self.api.binaries_meta.find_hashes([self.non_existent_sha256])
        self.assertEqual(1, len(result.items))
        self.assertFalse(result.items[0].exists)

    # -------------------------------- Term Query
    def test_find_term_query_basic(self):
        term = f'source.encoded_references.value:"{self.user}"'
        resp = self.api.binaries_meta.find(term)
        # Check in the module setup if this number needs to be increased/decreased.
        self.assertGreaterEqual(len(resp.items), self.default_min_hash_count)

    def _assert_sorting_source_timestamp(
        self, resp_newest_first: models_restapi.EntityFind, resp_oldest_first: models_restapi.EntityFind
    ):
        """Verifies that source_timestamp filtering is working appropriately.

        First checks at least one of the newest submissions are one of the ones subitted by the test setup.
        Then checks that the first item in the newest list has a larger source date than the oldest list.
        """
        # if sorting worked the reverse of one list equals the other
        resp_newest_first_sha256s = list([i.sha256 for i in resp_newest_first.items])

        # Verify at least one of the newest
        at_least_one_hash_found = False
        for newest_hash in resp_newest_first_sha256s:
            if newest_hash in self.sha256s:
                at_least_one_hash_found = True
        self.assertTrue(
            at_least_one_hash_found,
            "None of the newest submission hashes were found when sorting by "
            + f"newest submission first and these should be the newest submissions! [{', '.join(resp_newest_first_sha256s)}]",
        )
        newest = pendulum.parse(resp_newest_first.items[0].sources[0].timestamp)
        oldest = pendulum.parse(resp_oldest_first.items[0].sources[0].timestamp)
        self.assertGreaterEqual(newest, oldest)

    def test_find_term_query_sorting(self):
        term = f'source.encoded_references.value:"{self.user}"'
        resp_oldest_first = self.api.binaries_meta.find(
            term, sort_asc=True, sort_prop=models_restapi.FindBinariesSortEnum.source_timestamp
        )
        resp_newest_first = self.api.binaries_meta.find(
            term, sort_asc=False, sort_prop=models_restapi.FindBinariesSortEnum.source_timestamp
        )
        self._assert_sorting_source_timestamp(resp_newest_first, resp_oldest_first)

    def test_find_term_query_limiting(self):
        term = f'source.encoded_references.value:"{self.user}"'
        resp = self.api.binaries_meta.find(term, max_entities=2)
        # Check in the module setup if this number needs to be increased/decreased.
        self.assertGreaterEqual(len(resp.items), 2)

    def test_find_term_query_count(self):
        term = f'source.encoded_references.value:"{self.user}"'
        resp = self.api.binaries_meta.find(term, count_entities=True)
        self.assertIsNotNone(resp.items_count)

    # -------------------------------- Simple Query
    def test_find_simple_basic(self):
        find_options = FindOptions(source_username=self.user)
        resp = self.api.binaries_meta.find_simple(find_options)
        # Check in the module setup if this number needs to be increased/decreased.
        self.assertGreaterEqual(len(resp.items), self.default_min_hash_count)

    def test_find_simple_sorting(self):
        find_options = FindOptions(source_username=self.user)
        resp_oldest_first = self.api.binaries_meta.find_simple(
            find_options, sort_asc=True, sort_prop=models_restapi.FindBinariesSortEnum.source_timestamp
        )
        resp_newest_first = self.api.binaries_meta.find_simple(
            find_options, sort_asc=False, sort_prop=models_restapi.FindBinariesSortEnum.source_timestamp
        )
        self._assert_sorting_source_timestamp(resp_newest_first, resp_oldest_first)

    def test_find_simple_limit(self):
        find_options = FindOptions(source_username=self.user)
        resp = self.api.binaries_meta.find_simple(find_options, max_entities=2)
        # Check in the module setup if this number needs to be increased/decreased.
        self.assertGreaterEqual(len(resp.items), 2)

    def test_find_simple_count(self):
        find_options = FindOptions(source_username=self.user)
        resp = self.api.binaries_meta.find_simple(find_options, count_entities=True)
        self.assertIsNotNone(resp.items_count)

    def test_simple_compound_query(self):
        options = FindOptions(
            source_username=self.user,
            sources=self.source,
            source_timestamp_older=pendulum.now(),
            source_timestamp_newer=pendulum.datetime(2020, 1, 1, 1, 1, 1, 1, 1),
            source_depth=0,
        )
        self._search_count_ge(options, self.default_min_hash_count)

    # -------------------------------- Simple Query - extended
    def _check_doesnt_fail(self, options: FindOptions) -> models_restapi.EntityFind:
        self._search_count_ge(options, 0)

    # --- SOURCE ID
    def _search_count_ge(self, options: FindOptions, count: int) -> models_restapi.EntityFind:
        resp = self.api.binaries_meta.find_simple(options, max_entities=count + 1)
        self.assertGreaterEqual(len(resp.items), count)
        return resp

    def _search_count_equal(self, options: FindOptions, count: int):
        # Look for just enough values to fail.
        resp = self.api.binaries_meta.find_simple(options, max_entities=count + 1)
        self.assertEqual(len(resp.items), count)

    def _get_source_ids(self, resp: models_restapi.EntityFind) -> set[str]:
        source_list = set()
        for val in resp.items:
            if not val.sources:
                # Note - This shouldn't be needed as sources should always be populated, but sometimes they aren't.
                continue
            for src in val.sources:
                source_list.add(src.name)
        return source_list

    def test_find_simple_sources(self):
        # single
        find_options = FindOptions(sources=self.source)
        resp = self._search_count_ge(find_options, self.default_min_hash_count)

        self.assertIn(self.source, self._get_source_ids(resp))
        # multiple
        included_sources = [self.source, self.source1]
        find_options = FindOptions(sources=included_sources)
        resp2 = self._search_count_ge(find_options, self.default_min_hash_count)
        at_least_one_source_included = False
        for src in self._get_source_ids(resp2):
            if src in included_sources:
                at_least_one_source_included = True
        self.assertTrue(at_least_one_source_included)

    def test_find_simple_sources_exclude(self):
        # single
        find_options = FindOptions(source_excludes=self.source1)
        resp = self._search_count_ge(find_options, self.default_min_hash_count)
        self.assertNotIn(self.source1, self._get_source_ids(resp))

        # multiple
        excluded_sources = [self.source1, self.source2]
        find_options = FindOptions(source_excludes=excluded_sources)
        resp2 = self._search_count_ge(find_options, self.default_min_hash_count)
        for src in self._get_source_ids(resp2):
            self.assertNotIn(src, excluded_sources)

    # --- SOURCE DEPTH

    def _get_source_depths(self, resp: models_restapi.EntityFind, prefer_depths: list[int] = None) -> list[int]:
        depths = []
        for item in resp.items:
            if not item.sources:
                # Note - This shouldn't be needed as sources should always be populated.
                continue
            current_item_depths = []
            current_item_depths_filtered = []
            for src in item.sources:
                if src:  # Ignore ids with no sources
                    current_item_depths.append(src.depth)
                    # Only keep the depths we expect
                    if prefer_depths and src.depth in prefer_depths:
                        print(f"Preferring {src.depth}")
                        current_item_depths_filtered.append(src.depth)
            # If the filtered response has data we have the depths we expected and maybe more.
            if len(current_item_depths_filtered) > 0:
                depths += current_item_depths_filtered
            else:
                # filter has filtered out all results so go with the unfiltered result.
                depths += current_item_depths
                print(f"(probably an error) accepting non preferred with values {current_item_depths} {item.sha256}")
        return depths

    # Depth is very hard to filter by, because Assemblyline meta-data is mapped which has a source_depth of 1.
    # But isn't the preferred source event.
    # When searching for an event with depth greater than 0 Assemblyline events that have been mapped will have depth 1
    # So those binaries will be 'valid' however when the binary is selected the actual depth provided will be 0
    # Because that's the depth of the source event which makes it hard to verify anything is working.

    # def test_find_simple_sources_source_depth(self):
    #     find_options = FindOptions(source_depth=0)
    #     resp = self._search_count_ge(find_options, self.default_min_hash_count)
    #     self.assertIn(0, self._get_source_depths(resp, prefer_depths=[0]))

    #     included_depths = [1, 2, 3, 4, 5, 6]
    #     find_options = FindOptions(source_depth=included_depths)
    #     resp2 = self._search_count_ge(find_options, self.default_min_hash_count)
    #     for d in self._get_source_depths(resp2, prefer_depths=included_depths):
    #         self.assertIn(d, included_depths)

    # FUTURE - the filter technically works but there is no way to verify the result.
    # Because metastore prefers the numerically smallest depth if you look for depth > 1 and there is depth
    # 1,2,3 on a binary it will be found and show depth 1 because that is the smallest number.
    # def test_find_simple_sources_source_depth_exclude(self):
    #     find_options = FindOptions(source_depth_exclude=1)
    #     resp = self._search_count_ge(find_options, 1)
    #     self.assertNotIn(1, self._get_source_depths(resp, prefer_depths=[2, 3, 4, 5, 6, 7, 8, 9, 10]))

    #     excluded_depths = [1, 2, 3, 4, 5, 6, 7, 8, 9]
    #     find_options = FindOptions(source_depth_exclude=excluded_depths)
    #     resp = self._search_count_ge(find_options, self.default_min_hash_count)
    #     for d in self._get_source_depths(resp, prefer_depths=[0, 10]):
    #         self.assertNotIn(d, excluded_depths)

    # def test_find_simple_sources_source_depth_less_than(self):
    #     find_options = FindOptions(source_depth_less=1)
    #     resp = self.api.binaries_meta.find_simple(find_options, max_entities=self.default_min_hash_count)
    #     for d in self._get_source_depths(resp, prefer_depths=[0]):
    #         self.assertLess(d, 1)

    # def test_find_simple_sources_source_depth_greater(self):
    #     find_options = FindOptions(source_depth_greater=0)
    #     resp = self._search_count_ge(find_options, self.default_min_hash_count)
    #     for d in self._get_source_depths(resp, prefer_depths=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10]):
    #         self.assertGreater(d, 0)

    # --- SOURCE TIMESTAMP
    def _check_timestamp(self, find_options: FindOptions, newer: bool, potentially_equal: bool):
        self._search_count_ge(find_options, self.default_min_hash_count)

    def test_find_entity_newer_than(self):
        find_options = FindOptions(source_timestamp_newer=pendulum.datetime(2020, 1, 1, 1, 1, 1, 1, 1))
        self._search_count_ge(find_options, self.default_min_hash_count)

    def test_find_entity_newer_than_or_equal_to(self):
        find_options = FindOptions(source_timestamp_newer_or_equal=pendulum.datetime(2020, 1, 1, 1, 1, 1, 1, 1))
        self._search_count_ge(find_options, self.default_min_hash_count)

    def test_find_entity_older_than(self):
        find_options = FindOptions(source_timestamp_older=pendulum.now())
        self._search_count_ge(find_options, self.default_min_hash_count)

    def test_find_entity_older_than_or_equal_to(self):
        find_options = FindOptions(source_timestamp_older_or_equal=pendulum.now())
        self._search_count_ge(find_options, self.default_min_hash_count)

    # --- PLUGIN/AUTHOR
    def test_find_entity_with_features_from_plugin_name(self):
        find_options = FindOptions(plugin_name=self.common_plugin_name)
        self._search_count_ge(find_options, self.default_min_hash_count)

    def test_find_entity_with_features_from_plugin_version(self):
        find_options = FindOptions(plugin_version="1.1.1")
        # Any 200 response is acceptable because, there may be no plugin version 1.1.1.
        self._check_doesnt_fail(find_options)

    # --- FEATURES
    def test_find_entity_with_feature_name(self):
        find_options = FindOptions(has_feature_keys=self.common_feature_key)
        self._search_count_ge(find_options, self.default_min_hash_count)

    def test_find_entity_with_feature_value(self):
        find_options = FindOptions(has_feature_values=self.common_feature_value)
        self._search_count_ge(find_options, self.default_min_hash_count)

    # --- BINARY INFO - ENTITY SIZE
    def test_find_entity_size_greater(self):
        find_options = FindOptions(greater_than_size_bytes=10)
        self._search_count_ge(find_options, self.default_min_hash_count)

    def test_find_entity_size_smaller(self):
        find_options = FindOptions(less_than_size_bytes=20000)
        self._search_count_ge(find_options, self.default_min_hash_count)

    def test_find_simple_file_format(self):
        # Normal
        find_options = FindOptions(file_formats=self.common_file_format)
        resp = self._search_count_ge(find_options, self.default_min_hash_count)
        for item in resp.items:
            self.assertEqual(self.common_file_format, item.file_format)
        # List
        find_options = FindOptions(file_formats=[self.common_file_format])
        resp = self._search_count_ge(find_options, self.default_min_hash_count)
        for item in resp.items:
            self.assertEqual(self.common_file_format, item.file_format)

    def test_find_simple_file_format_exclude(self):
        # Normal
        find_options = FindOptions(file_formats_exclude=self.common_file_format)
        resp = self._search_count_ge(find_options, self.default_min_hash_count)
        for item in resp.items:
            self.assertNotEqual(self.common_file_format, item.file_format)
        # List
        find_options = FindOptions(file_formats_exclude=[self.common_file_format])
        resp = self._search_count_ge(find_options, self.default_min_hash_count)
        for item in resp.items:
            self.assertNotEqual(self.common_file_format, item.file_format)

    def test_find_all(self):
        # should have at least 400 binaries of any kind
        found = 0
        find_options = FindOptions()
        find_result = self.api.binaries_meta.find_all(find_options, request_binaries=100, max_binaries=400)
        for item in find_result.iter:
            self.assertGreaterEqual(find_result.approx_total, 400)
            self.assertIsNotNone(item)
            found += 1
            if found > 400:
                break
        self.assertGreaterEqual(found, 400)

        # should have at least 40 binaries of common kind
        found = 0
        find_options = FindOptions(file_formats_exclude=self.common_file_format)
        find_result = self.api.binaries_meta.find_all(find_options, request_binaries=100, max_binaries=400)
        for item in find_result.iter:
            self.assertGreaterEqual(find_result.approx_total, 40)
            self.assertIsNotNone(item)
            found += 1
            if found > 40:
                break
        self.assertGreaterEqual(found, 40)

        # Should still be fine if there are no binaries that meet condition
        find_result = self.api.binaries_meta.find_all(
            FindOptions(greater_than_size_bytes=1000000000000000), request_binaries=100, max_binaries=1
        )
        for item in find_result:
            self.assertTrue(False, "Shouldn't iterate over a length zero iterator")
        self.assertEqual(find_result.approx_total, 0)

    # --- TAGS
    # FUTURE - can't test this search functionality because we need to be able to add a tag to a binary/feature to test the filter.
    # def test_find_simple_binary_tag(self):
    #     # At least a 200
    #     find_options = FindOptions(binary_tags=self.dummy_tag)
    #     self._check_doesnt_fail(find_options)

    # def test_find_simple_feature_tag(self):
    #     # At least a 200
    #     find_options = FindOptions(feature_tags=self.dummy_tag)
    #     self._check_doesnt_fail(find_options)



import json

from azul_bedrock import models_restapi

from azul_client.exceptions import BadResponse, BadResponse404

from .base_test import BaseApiTest


class TestSourcesApi(BaseApiTest):
    @classmethod
    def setUpClass(cls):
        cls.common_feature_magic = "magic"
        cls.common_feature_file_extension = "file_extension"
        cls.entropy_feature = "entropy"
        # There should be entropy of 0 in the system somewhere
        cls.known_magic_value = "ASCII text, with CRLF line terminators"
        cls.known_entropy_value = "0"
        cls.known_extension_feat_val = "txt"
        # Minimum number of values you expected a large feature like the above three to have.
        cls.large_feature_minimum_values = 10
        cls.entropy_xpart = "integer"

        # tagging
        cls.feature_tag_entropy = "test-fv-tag-1"
        cls.feature_tag_extension = "test-fv-tag-2"
        cls.feature_tag_magic = "test-fv-tag-3"
        super().setUpClass()

        # Create two feature_value tags (Note this is tested in this class but also needed to test this class.fd)
        tag_created_1 = cls.api.features.create_feature_value_tag(
            tag=cls.feature_tag_entropy,
            feature=cls.entropy_feature,
            value=cls.known_entropy_value,
            security=cls.default_security,
        )
        assert tag_created_1 is None, "Failed to setup for feature test, does Entropy have a single 0 value?"
        tag_created_2 = cls.api.features.create_feature_value_tag(
            tag=cls.feature_tag_extension,
            feature=cls.common_feature_file_extension,
            value=cls.known_extension_feat_val,
            security=cls.default_security,
        )
        assert tag_created_2 is None, (
            "Failed to setup for feature test, is there at least one file with a txt extension."
        )

        # Find the current version of Entropy for later use.
        list_of_plugins = cls.api.plugins.get_all_plugins()
        current_version_of_entropy = None
        for plugin in list_of_plugins:
            if plugin.newest_version.name == cls.entropy_plugin_name:
                current_version_of_entropy = plugin.newest_version.version
                break
        if current_version_of_entropy is None:
            raise Exception(
                "Something is wrong with the plugins API or Entropy, can't find latest version of entropy."
            )
        cls.entropy_version = current_version_of_entropy

    def test_count_unique_values_in_feature(self):
        """Test counting the number of unique feature/value combinations there are for a given feature."""
        empty_counts = self.api.features.count_unique_values_in_feature([])
        self.assertEqual(empty_counts, dict())

        # MAGIC FEATURE COUNT ---
        magic_count = self.api.features.count_unique_values_in_feature([self.common_feature_magic])
        self.assertEqual(len(magic_count.keys()), 1)
        self.assertIn(self.common_feature_magic, magic_count.keys())
        count_val = magic_count[self.common_feature_magic]
        self.assertEqual(count_val.name, self.common_feature_magic)
        # There should be at least 20 magic values as there will be hundreds or thousands on a large system.
        self.assertGreaterEqual(count_val.values, self.large_feature_minimum_values)

        # MAGIC AND FILE EXTENSION FEATURE COUNT ---
        multi_count = self.api.features.count_unique_values_in_feature(
            [self.common_feature_magic, self.common_feature_file_extension]
        )
        self.assertGreaterEqual(len(multi_count.keys()), 2)
        self.assertIn(self.common_feature_magic, multi_count.keys())
        self.assertIn(self.common_feature_file_extension, multi_count.keys())

        # ENTROPY with just author name ---
        entropy_response = self.api.features.count_unique_values_in_feature(
            [self.entropy_feature], author=self.entropy_plugin_name
        )
        self.assertEqual(list(entropy_response.keys()), [self.entropy_feature])
        # Should be at least 20 entropies calculated ---
        self.assertGreaterEqual(entropy_response[self.entropy_feature].values, self.large_feature_minimum_values)

        # ENTROPY with author name and version ---
        entropy_response_2 = self.api.features.count_unique_values_in_feature(
            [self.entropy_feature], author=self.entropy_plugin_name, author_version=self.entropy_version
        )
        self.assertEqual(list(entropy_response_2.keys()), [self.entropy_feature])
        # Should be at least 20 entropies calculated
        self.assertGreaterEqual(entropy_response_2[self.entropy_feature].values, self.large_feature_minimum_values)

    def test_count_unique_entities_in_features(self):
        """Test counting the number of entities(binaries) that have a particular feature."""
        empty_counts = self.api.features.count_unique_values_in_feature([])
        self.assertEqual(empty_counts, dict())

        # MAGIC FEATURE COUNT ---
        magic_count = self.api.features.count_unique_entities_in_features([self.common_feature_magic])
        self.assertEqual(len(magic_count.keys()), 1)
        self.assertIn(self.common_feature_magic, magic_count.keys())
        count_val = magic_count[self.common_feature_magic]
        self.assertEqual(count_val.name, self.common_feature_magic)
        # There should be at least 20 entities with magic as there will be hundreds or thousands on a large system.
        self.assertGreaterEqual(count_val.entities, self.large_feature_minimum_values)

        # MAGIC AND FILE EXTENSION FEATURE COUNT ---
        multi_count = self.api.features.count_unique_entities_in_features(
            [self.common_feature_magic, self.common_feature_file_extension]
        )
        self.assertGreaterEqual(len(multi_count.keys()), 2)
        self.assertIn(self.common_feature_magic, multi_count.keys())
        self.assertIn(self.common_feature_file_extension, multi_count.keys())

        # ENTROPY with just author name ---
        entropy_response = self.api.features.count_unique_entities_in_features(
            [self.entropy_feature], author=self.entropy_plugin_name
        )
        self.assertEqual(list(entropy_response.keys()), [self.entropy_feature])
        # Should be at least 20 entropies calculated ---
        self.assertGreaterEqual(entropy_response[self.entropy_feature].entities, self.large_feature_minimum_values)

        # ENTROPY with author name and version ---
        entropy_response_2 = self.api.features.count_unique_entities_in_features(
            [self.entropy_feature], author=self.entropy_plugin_name, author_version=self.entropy_version
        )
        self.assertEqual(list(entropy_response_2.keys()), [self.entropy_feature])
        # Should be at least 20 entropies calculated
        self.assertGreaterEqual(entropy_response_2[self.entropy_feature].entities, self.large_feature_minimum_values)

    def test_count_unique_entities_in_featurevalues(self):
        """Test counting the number of entities(binaries) that have a particular feature/value pair."""
        resp = self.api.features.count_unique_entities_in_featurevalues([])
        self.assertEqual(resp, dict())

        # Request with feature and value no AUTHOR ---
        resp = self.api.features.count_unique_entities_in_featurevalues(
            [models_restapi.ValueCountItem(name=self.entropy_feature, value=self.known_entropy_value)]
        )
        self.assertEqual(list(resp.keys()), [self.entropy_feature])
        self.assertEqual(list(resp[self.entropy_feature].keys())[0], self.known_entropy_value)
        val_entropy_feature_count = resp[self.entropy_feature][self.known_entropy_value]
        self.assertEqual(val_entropy_feature_count.value, self.known_entropy_value)
        self.assertGreater(val_entropy_feature_count.entities, self.large_feature_minimum_values)

        # Request with feature and value with AUTHOR ---
        resp = self.api.features.count_unique_entities_in_featurevalues(
            [models_restapi.ValueCountItem(name=self.entropy_feature, value=self.known_entropy_value)],
            author=self.entropy_plugin_name,
        )
        self.assertEqual(list(resp.keys()), [self.entropy_feature])
        self.assertEqual(list(resp[self.entropy_feature].keys())[0], self.known_entropy_value)
        val_entropy_feature_count = resp[self.entropy_feature][self.known_entropy_value]
        self.assertEqual(val_entropy_feature_count.value, self.known_entropy_value)
        self.assertGreater(val_entropy_feature_count.entities, self.large_feature_minimum_values)

        # Request with feature and value with AUTHOR AND VERSION ---
        resp = self.api.features.count_unique_entities_in_featurevalues(
            [models_restapi.ValueCountItem(name=self.entropy_feature, value=self.known_entropy_value)],
            author=self.entropy_plugin_name,
            author_version=self.entropy_version,
        )
        self.assertEqual(list(resp.keys()), [self.entropy_feature])
        self.assertEqual(list(resp[self.entropy_feature].keys())[0], self.known_entropy_value)
        val_entropy_feature_count = resp[self.entropy_feature][self.known_entropy_value]
        self.assertEqual(val_entropy_feature_count.value, self.known_entropy_value)
        self.assertGreater(val_entropy_feature_count.entities, self.large_feature_minimum_values)

    def test_count_unique_entities_in_featurevalueparts(self):
        """Test the part counts of number of binaries/entities with a given feature value."""
        resp = self.api.features.count_unique_entities_in_featurevalues([])
        self.assertEqual(resp, dict())

        # Request with feature and value no AUTHOR ---
        # expected form {'0': {'integer': models.ValuePartCountRet(value='0', part='integer', entities=3025848)}}
        resp = self.api.features.count_unique_entities_in_featurevalueparts(
            [models_restapi.ValuePartCountItem(part=self.entropy_xpart, value=self.known_entropy_value)]
        )
        self.assertEqual(list(resp.keys()), [self.known_entropy_value])
        found_value = resp[self.known_entropy_value]
        self.assertEqual(list(found_value.keys()), [self.entropy_xpart])
        self.assertEqual(found_value[self.entropy_xpart].value, self.known_entropy_value)
        self.assertGreater(found_value[self.entropy_xpart].entities, self.large_feature_minimum_values)

        # Request with feature and value with AUTHOR ---
        resp = self.api.features.count_unique_entities_in_featurevalueparts(
            [models_restapi.ValuePartCountItem(part=self.entropy_xpart, value=self.known_entropy_value)],
            author=self.entropy_plugin_name,
        )
        self.assertEqual(list(resp.keys()), [self.known_entropy_value])
        found_value = resp[self.known_entropy_value]
        self.assertEqual(list(found_value.keys()), [self.entropy_xpart])
        self.assertEqual(found_value[self.entropy_xpart].value, self.known_entropy_value)
        self.assertEqual(found_value[self.entropy_xpart].entities, 0)

        # Request with feature and value with AUTHOR AND VERSION ---
        resp = self.api.features.count_unique_entities_in_featurevalueparts(
            [models_restapi.ValuePartCountItem(part=self.entropy_xpart, value=self.known_entropy_value)],
            author=self.entropy_plugin_name,
            author_version=self.entropy_version,
        )
        self.assertEqual(list(resp.keys()), [self.known_entropy_value])
        found_value = resp[self.known_entropy_value]
        self.assertEqual(list(found_value.keys()), [self.entropy_xpart])
        self.assertEqual(found_value[self.entropy_xpart].value, self.known_entropy_value)
        self.assertEqual(found_value[self.entropy_xpart].entities, 0)

    def test_get_all_feature_value_tags(self):
        resp = self.api.features.get_all_feature_value_tags()
        self.assertGreaterEqual(resp.num_tags, 1)
        self.assertEqual(len(resp.tags), resp.num_tags)

    def test_get_feature_values_in_tag(self):
        """Test getting a list of all features and values for a given tag."""
        resp = self.api.features.get_feature_values_in_tag(self.feature_tag_entropy)
        is_entropy_fv_found = False
        for feature_val_tag in resp.items:
            # Verify we've found the expected feature/value tagged.
            if (
                feature_val_tag.feature_name == self.entropy_feature
                and feature_val_tag.feature_value == self.known_entropy_value
            ):
                is_entropy_fv_found = True

        self.assertTrue(
            is_entropy_fv_found,
            f"Entropy's known value '{self.known_entropy_value}' wasn't found with the appropriate tag,"
            + f" fv with tag were {resp.items}",
        )

    def test_create_feature_value_tag(self):
        resp = self.api.features.create_feature_value_tag(
            self.feature_tag_magic, self.common_feature_magic, self.known_magic_value, self.default_security
        )
        self.assertIsNone(resp)

        # Creation with bad security.
        with self.assertRaises(BadResponse):
            self.api.features.create_feature_value_tag(
                self.feature_tag_magic, self.common_feature_magic, self.known_magic_value, "SJDKFNIEJUNIJNSEJCDN D"
            )

    def test_delete_feature_value_tag(self):
        def _verify_feature_value():
            found = False
            resp = self.api.features.get_feature_values_in_tag(self.feature_tag_extension)
            for fv in resp.items:
                if (
                    fv.feature_name == self.common_feature_file_extension
                    and fv.feature_value == self.known_extension_feat_val
                ):
                    found = True
                    break
            return found

        self.assertTrue(
            _verify_feature_value(),
            f"Feature '{self.common_feature_file_extension}' with expected value "
            + f"'{self.known_extension_feat_val}' does not exist to delete.",
        )
        self.api.features.delete_feature_value_tag(
            self.feature_tag_extension, self.common_feature_file_extension, self.known_extension_feat_val
        )
        self.assertFalse(
            _verify_feature_value(),
            f"Feature '{self.common_feature_file_extension}' with expected value "
            + f"'{self.known_extension_feat_val}' was not deleted.",
        )

    def test_find_features(self):
        resp = self.api.features.find_features()
        self.assertGreater(len(resp.items), self.large_feature_minimum_values)
        resp = self.api.features.find_features(author=self.entropy_plugin_name)
        self.assertGreaterEqual(len(resp.items), 1)
        resp = self.api.features.find_features(author=self.entropy_plugin_name, author_version=self.entropy_version)
        self.assertGreaterEqual(len(resp.items), 1)
        self.assertIn(self.entropy_feature, [fv.name for fv in resp.items])
        # No items as author is invalid.
        resp = self.api.features.find_features(author="jakldfjaklvnewiourjnsd")
        self.assertEqual(len(resp.items), 0)

    def test_find_values_in_feature_minimal(self):
        resp = self.api.features.find_values_in_feature(self.entropy_feature)
        self.assertEqual(resp.name, self.entropy_feature)
        self.assertEqual(resp.type, "float")
        self.assertGreater(len(resp.values), self.large_feature_minimum_values)
        self.assertIsNotNone(resp.after)

    def test_find_values_in_feature_with_term(self):
        """Search for the start of the known feature value and it should be found in the term query."""
        resp = self.api.features.find_values_in_feature(
            self.common_feature_file_extension, term=self.known_extension_feat_val[0]
        )
        self.assertEqual(resp.name, self.common_feature_file_extension)
        self.assertEqual(resp.type, "string")
        self.assertGreaterEqual(len(resp.values), 1)
        self.assertGreaterEqual(resp.total, 1)

        self.assertIn(self.known_extension_feat_val, [val.value for val in resp.values])

    def test_find_values_in_feature_with_author_version(self):
        """Search with author and sorts set."""
        resp = self.api.features.find_values_in_feature(
            self.entropy_feature, sort_asc=False, case_insensitive=True, author_version=self.entropy_version
        )
        self.assertEqual(resp.name, self.entropy_feature)
        self.assertEqual(resp.type, "float")
        self.assertGreater(len(resp.values), self.large_feature_minimum_values)
        self.assertGreater(len(resp.after if resp.after else ""), 1)



import time
import urllib
from unittest import mock

import pytest
from pytest_httpx import HTTPXMock

from azul_client import Api, Config, client
from azul_client import config as mconfig
from azul_client.oidc import OIDC

jwt = "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJzdWIiOiIxMjM0NTY3ODkwIiwibmFtZSI6IkpvaG4gRG9lIiwiaWF0IjoxNTE2MjM5MDIyfQ.SflKxwRJSMeKKF2QT4fwpMeJf36POk6yJV_adQssw5c"
OIDC_URL = "http://localhost:8123"


@pytest.fixture
def api_service_api() -> Api:
    """Fixture container the initialised api."""
    mconfig.location = ""
    cur_api = Api(
        mconfig.Config(
            auth_type="service",
            oidc_url=OIDC_URL,
            auth_client_id="servicer",
            auth_client_secret="secrot",
        )
    )
    client.api = cur_api
    return cur_api


def test_via_service_token(api_service_api: Api, httpx_mock: HTTPXMock):
    endpt = api_service_api.config.oidc_url
    print(f"{endpt=}")

    wellknown = {
        "authorization_endpoint": f"{endpt}/authz",
        "token_endpoint": f"{endpt}/token",
    }
    httpx_mock.add_response(
        method="GET",
        url=endpt,
        json=wellknown,
        headers={"content-type": "application/json; blarg"},
    )
    httpx_mock.add_response(
        method="POST",
        url=endpt + "/token",
        json={"access_token": "me give access"},
        headers={"content-type": "application/json; blarg"},
    )
    token = api_service_api.auth._via_service_token()

    last_request = httpx_mock.get_requests()[-1]
    data = last_request.read()
    parsed_content = urllib.parse.parse_qs(data.decode())
    assert "me give access" == token["access_token"]
    assert "client_credentials" == parsed_content["grant_type"][0]
    assert "secrot" == parsed_content["client_secret"][0]
    assert "servicer" == parsed_content["client_id"][0]


@pytest.fixture
def api_code_ex_api() -> Api:
    """Fixture container the initialised api."""
    mconfig.location = ""
    cur_api = Api(
        mconfig.Config(
            auth_type="callback",
            oidc_url=OIDC_URL,
            auth_client_id="me",
        )
    )
    client.api = cur_api
    return cur_api


def test_via_code_callback(api_code_ex_api: Api, httpx_mock: HTTPXMock):
    endpt = api_code_ex_api.config.oidc_url
    wellknown = {
        "authorization_endpoint": f"{endpt}/authz",
        "token_endpoint": f"{endpt}/token",
    }
    httpx_mock.add_response(
        method="GET",
        url=endpt,
        json=wellknown,
        headers={"content-type": "application/json; blarg"},
    )
    httpx_mock.add_response(
        method="POST",
        url=endpt + "/token",
        json={"access_token": jwt},
        headers={"content-type": "application/json; blarg"},
    )
    with mock.patch("azul_client.oidc.callback.receive_code", lambda *k, **v: "the_code"):
        token = api_code_ex_api.auth._via_code_callback()

    last_request = httpx_mock.get_requests()[-1]
    data = last_request.read()
    parsed_content = urllib.parse.parse_qs(data.decode())
    assert jwt == token["access_token"]
    assert "authorization_code" == parsed_content["grant_type"][0]
    assert "me" == parsed_content["client_id"][0]


@pytest.fixture
def oidc_svc_callback() -> OIDC:
    mconfig.location = ""
    return OIDC(
        Config(
            auth_type="callback",
            oidc_url=OIDC_URL,
            auth_client_id="servicer",
        )
    )


def _mock_save(x):
    pass


@mock.patch("azul_client.oidc.oidc.OIDC._via_code_callback")
@mock.patch("azul_client.oidc.oidc.OIDC._via_refresh")
@mock.patch("azul_client.config.Config.save", _mock_save)
def test_get_token(vr: mock.MagicMock, vcc: mock.MagicMock, oidc_svc_callback: OIDC):
    vcc.return_value = {"access_token": "token"}
    vr.return_value = {"access_token": "refreshed_token"}
    # trigger initial auth
    assert "token" == oidc_svc_callback._get_access_token()
    assert 1 == vcc.call_count
    assert 0 == vr.call_count
    # reuse existing token
    assert "token" == oidc_svc_callback._get_access_token()
    assert 1 == vcc.call_count
    assert 0 == vr.call_count
    # check that refresh token is used
    oidc_svc_callback.cfg.auth_token_time = time.time() - 60
    assert "refreshed_token" == oidc_svc_callback._get_access_token()
    assert 1 == vcc.call_count
    assert 1 == vr.call_count
    # pretend that refresh token expired, so full auth required
    vr.return_value = None
    oidc_svc_callback.cfg.auth_token_time = time.time() - 600
    assert "token" == oidc_svc_callback._get_access_token()
    assert 2 == vcc.call_count
    assert 2 == vr.call_count


@mock.patch("azul_client.oidc.oidc.OIDC._via_code_callback")
@mock.patch("azul_client.config.Config.save", _mock_save)
def test_get_token_errors(vcc: mock.MagicMock, oidc_svc_callback: OIDC, httpx_mock: HTTPXMock):
    endpt = oidc_svc_callback.cfg.oidc_url
    vcc.return_value = {"access_token": "token", "refresh_token": "invalid"}
    # trigger initial auth
    assert "token" == oidc_svc_callback._get_access_token()
    assert 1 == vcc.call_count
    # reuse existing token
    assert "token" == oidc_svc_callback._get_access_token()
    assert 1 == vcc.call_count
    # check that refresh token failure 400 results in full reauth
    vcc.return_value = {"access_token": "redo_token", "refresh_token": "invalid"}
    wellknown = {
        "authorization_endpoint": f"{endpt}/authz",
        "token_endpoint": f"{endpt}/token",
    }
    httpx_mock.add_response(
        method="GET",
        url=endpt,
        json=wellknown,
        headers={"content-type": "application/json; blarg"},
    )
    httpx_mock.add_response(
        method="POST",
        url=endpt + "/token",
        content="refresh token expired",
        headers={"content-type": "application/json; blarg"},
        status_code=400,
    )
    oidc_svc_callback.cfg.auth_token_time = time.time() - 60
    assert "redo_token" == oidc_svc_callback._get_access_token()

    vcc.return_value = {"access_token": "redo_token2", "refresh_token": "invalid"}
    httpx_mock.add_response(
        method="POST",
        url=endpt + "/token",
        content="refresh token expired",
        headers={"content-type": "application/json; blarg"},
        status_code=401,
    )
    oidc_svc_callback.cfg.auth_token_time = time.time() - 60
    assert "redo_token2" == oidc_svc_callback._get_access_token()



from azul_client.exceptions import BadResponse, BadResponse404

from .base_test import BaseApiTest


class TestSecurityApi(BaseApiTest):
    def test_get_security_settings(self):
        """Verify security settings are a dictionary and have some basic keys."""
        resp = self.api.security.get_security_settings()
        print(resp)
        self.assertIsInstance(resp, dict)
        # Verify some of the keys you would expect to find in security settings are there.
        self.assertIn("labels", resp)
        self.assertIn("default", resp)

    def test_normalise_basic(self):
        """Minimal tests in case security changes."""
        resp_content = self.api.security.normalise("OFFICIAL      TLP:GREEN")
        self.assertEqual(resp_content, "OFFICIAL TLP:GREEN")

        resp_content = self.api.security.normalise("OFFICIAL TLP:GREEN")
        self.assertEqual(resp_content, "OFFICIAL TLP:GREEN")

        resp_content = self.api.security.normalise("OFFICIAL//TLP:GREEN")
        self.assertEqual(resp_content, "OFFICIAL TLP:GREEN")

        resp_content = self.api.security.normalise("LESS OFFICIAL//TLP:CLEAR")
        self.assertEqual(resp_content, "LESS OFFICIAL TLP:CLEAR")

        resp_content = self.api.security.normalise("MORE OFFICIAL//REL:APPLE")
        self.assertEqual(resp_content, "MORE OFFICIAL REL:APPLE")

        self.assertRaises(BadResponse, self.api.security.normalise, "OFFICIAL//TLP:GREEN//REL:APPLE")
        self.assertRaises(BadResponse, self.api.security.normalise, "OFFICIAL//REL:APPLE")
        self.assertRaises(BadResponse, self.api.security.normalise, "OFFICIAL AMBER")
        self.assertRaises(BadResponse, self.api.security.normalise, "OFF")
        self.assertRaises(BadResponse, self.api.security.normalise, "TLP:GREEN")
        self.assertRaises(BadResponse, self.api.security.normalise, "REL:APPLE")

    def test_get_max_security_string(self):
        max_security = self.api.security.get_max_security_string(["OFFICIAL TLP:GREEN"])
        self.assertEqual(max_security, "OFFICIAL TLP:GREEN")
        max_security = self.api.security.get_max_security_string(
            ["OFFICIAL TLP:CLEAR", "OFFICIAL TLP:GREEN", "OFFICIAL TLP:AMBER"]
        )
        self.assertEqual(max_security, "OFFICIAL TLP:AMBER")
        max_security = self.api.security.get_max_security_string(["MORE OFFICIAL REL:APPLE"])
        self.assertEqual(max_security, "MORE OFFICIAL REL:APPLE")
        with self.assertRaises(Exception):
            self.api.security.get_max_security_string("OFFICIAL//TLP:CLEAR")

        with self.assertRaises(Exception):
            self.api.security.get_max_security_string("REL:APPLE")

    def test_get_is_user_an_admin(self):
        """Verify the test user is an admin user."""
        is_user_admin = self.api.security.get_is_user_an_admin()
        self.assertIsInstance(is_user_admin, bool)
        self.assertTrue(is_user_admin)



"""Module for interacting with binary endpoints."""

import logging
from dataclasses import dataclass
from http import HTTPMethod
from typing import Callable, Generator

import httpx
import pendulum
from azul_bedrock import models_network as azm
from azul_bedrock import models_restapi
from pydantic import TypeAdapter

from azul_client import config
from azul_client.api.base_api import BaseApiHandler

logger = logging.getLogger(__name__)
autocomplete_type_adapter = TypeAdapter(models_restapi.AutocompleteContext)


@dataclass
class FindOptions:
    """Store all of the options available for creating a term query.

    Note list values assume an OR relationship when looking for matches.
    Note list values assume an AND relationship when excluding matches.

    Note dict values always assume an AND relationship between members of dict.

    Note when specifying string values you can provide a wildcard '*' character.
    Typically only useful at the end, rather than the start or middle e.g when looking for al file type image/bmp:
    works:
    file_format:"image*"
    doesn't work:
    file_format:"ima*e"
    file_format:"*bmp"
    """

    _query: str = ""

    # ------------------------- Source options
    # Name(id) of the source(s) to look for entities within.
    sources: list[str] | str | None = None
    # Name(id) of the source(s) to exclude when looking for entities.
    source_excludes: list[str] | str | None = None

    # Includes only entities that have the depths provided.
    source_depth: list[int] | int | None = None
    # Exclude entities that have the depths provided.
    source_depth_exclude: list[int] | int | None = None
    # Includes only entities that have the depth greater than the value provided.
    source_depth_greater: int | None = None
    # Includes only entities that have the depth less than the value provided.
    source_depth_less: int | None = None

    # Includes only entities that have been sourced by the username.
    source_username: str | None = None
    # Includes only entities that have ALL the provided reference fields.
    source_reference: dict[str, str] | None = None  # Key value pairs of reference fields.

    # ------------------------- Time options
    # Include only entities that are newer than or equal to the provided timestamp (to nearest second)
    source_timestamp_newer_or_equal: pendulum.DateTime | None = None
    # Include only entities that are older than or equal to the provided timestamp (to nearest second)
    source_timestamp_older_or_equal: pendulum.DateTime | None = None
    # Include only entities that are newer than the provided timestamp (to nearest second)
    source_timestamp_newer: pendulum.DateTime | None = None
    # Include only entities that are older than the provided timestamp (to nearest second)
    source_timestamp_older: pendulum.DateTime | None = None

    # ------------------------- Plugin/Author filtering
    # Include entities that have valid results from the provided plugin (case sensitive).
    plugin_name: str | None = None
    # Include entities that have valid results from the provided plugin version (case sensitive).
    plugin_version: str | None = None

    # ------------------------- Feature filtering
    # Include entities that have any of the feature keys.
    has_feature_keys: list[str] | str | None = None
    # Include entities that have any of the feature values.
    has_feature_values: list[str] | str | None = None

    # ------------------------- Binary Info
    # Include entities that are greater (in bytes) than the provided value.
    greater_than_size_bytes: int | None = None
    # Include entities that are less than (in bytes) than the provided value.
    less_than_size_bytes: int | None = None

    # Include entities that have the specified file type.
    file_formats_legacy: str | list[str] | None = None
    # Exclude entities that have the specified file type.
    file_formats_legacy_exclude: str | list[str] | None = None
    # Include entities that have the specified file type (AL type).
    file_formats: str | list[str] | None = None
    # Exclude entities that have the specified file type (AL type).
    file_formats_exclude: str | list[str] | None = None

    # ------------------------- Tags
    # Find all binaries that have any of the provided tags.
    binary_tags: str | list[str] | None = None
    # Final all binaries that have features with the provided tags.
    feature_tags: str | list[str] | None = None

    # --- End of options ---

    def _add(self, value: str):
        """Append a new value to the query."""
        if not self._query:
            self._query = value
        else:
            self._query += " " + value

    def _add_date(self, search_key: str, value: pendulum.DateTime | None):
        """Add a date as an integer to the query."""
        if value is None:
            return
        # Timestamp is just milliseconds since epoch, Format docs - https://pendulum.eustace.io/docs/#string-formatting
        self._add(search_key % (value.format("x")))

    def _add_if_not_none(self, search_key: str, value: str | int | None):
        """Add a value to the internal query if the provided value isn't none."""
        if value is None:
            return
        self._add(search_key % (value))

    def _add_list(self, search_key: str, value: str | list[str] | int | list[int] | None, negation: bool = False):
        """Add a list of values to the internal query if the provided value isn't none.

        Negation is used to switch between 'AND'ing and 'OR'ing members of the list.
        """
        if value is None or (isinstance(value, list) and len(value) == 0):
            return

        # Convert to a list if required.
        if isinstance(value, str) or isinstance(value, int):
            value = [value]

        for i, val in enumerate(value):
            if i > 0:
                prefix = "OR "
                if negation:
                    prefix = "AND "
                self._add(prefix + search_key % (val))
            else:
                self._add(search_key % (val))

    def _add_key_value(self, search_key: str, value: dict[str, str] | None):
        """Add a number key value pair queries."""
        if value is None or len(value) == 0:
            return

        for k, v in value.items():
            self._add(search_key % (k, v))

    def to_query(self) -> str:
        """Convert the find options into a term query."""
        self._query = ""
        # Source options
        self._add_list('source.name:"%s"', self.sources)
        self._add_list('!source.name:"%s"', self.source_excludes, negation=True)

        self._add_list("depth:%s", self.source_depth)
        self._add_list("!depth:%s", self.source_depth_exclude, negation=True)
        self._add_if_not_none("depth:>%s", self.source_depth_greater)
        self._add_if_not_none("depth:<%s", self.source_depth_less)

        self._add_if_not_none('source.encoded_references.key_value:"user.%s"', self.source_username)
        self._add_key_value('source.encoded_references.key_value:"%s.%s"', self.source_reference)

        # Time options
        self._add_date("source.timestamp:>=%s", self.source_timestamp_newer_or_equal)
        self._add_date("source.timestamp:<=%s", self.source_timestamp_older_or_equal)
        self._add_date("source.timestamp:>%s", self.source_timestamp_newer)
        self._add_date("source.timestamp:<%s", self.source_timestamp_older)

        # Author options
        self._add_if_not_none('author.name:"%s"', self.plugin_name)
        self._add_if_not_none('author.version:"%s"', self.plugin_version)

        # Features
        self._add_list('features.name:"%s"', self.has_feature_keys)
        self._add_list('features.value:"%s"', self.has_feature_values)

        # Binary Info
        self._add_if_not_none("size:>%s", self.greater_than_size_bytes)
        self._add_if_not_none("size:<%s", self.less_than_size_bytes)

        self._add_list('file_format:"%s"', self.file_formats)
        self._add_list('!file_format:"%s"', self.file_formats_exclude, negation=True)

        # Tags
        self._add_list('binary.tag:"%s"', self.binary_tags)
        self._add_list('feature.tag:"%s"', self.feature_tags)

        return self._query


class BinariesMeta(BaseApiHandler):
    """Interact with binary endpoints."""

    SHA256_regex = r"^[a-fA-F0-9]{64}$"
    upload_download_timeout = 120

    def __init__(self, cfg: config.Config, get_client: Callable[[], httpx.Client]) -> None:
        """Init."""
        super().__init__(cfg, get_client)

    def check_meta(self, sha256: str) -> bool:
        """Check metadata exists for hash."""
        return self._generic_head_request(self.cfg.azul_url + f"/api/v0/binaries/{sha256}")

    def get_meta(
        self,
        sha256: str,
        *,
        details: list[models_restapi.BinaryMetadataDetail] | None = None,
        author: str | None = None,
        bucket_size: int = 100,
    ) -> models_restapi.BinaryMetadata:
        """Get metadata for hash.

        :param str sha256: sha256 of the binary to get metadata for.
        :param bool detail: Set to True to get detailed information about the binary including
        children, features, streams, parents etc.
        :param int bucket_size: Edit bucket size to get data if a query overflows the current bucket count
        (Buckets this affects are Features, Info, Streams(data) and Instances(Authors)).
        """
        params = {"detail": details, "author": author, "bucket_size": bucket_size}
        for k in list(params.keys()):
            if params[k] is None:
                params.pop(k)
        return self._request_with_pydantic_model_response(
            method=HTTPMethod.GET,
            url=self.cfg.azul_url + f"/api/v0/binaries/{sha256}",
            response_model=models_restapi.BinaryMetadata,
            params=params,
            get_data_only=True,
        )

    def _base_find(
        self,
        *,
        term: str | None = None,
        max_entities: int | None = None,
        count_entities: bool | None = None,
        hashes: list[str] | None = None,
        sort_prop: models_restapi.FindBinariesSortEnum | None = None,
        sort_asc: bool | None = None,
    ) -> models_restapi.EntityFind:
        params = dict(
            term=term,
            max_entities=max_entities,
            count_entities=count_entities,
            sort=sort_prop,
            sort_asc=sort_asc,
        )
        params = self.filter_none_values(params)

        if not hashes:
            hashes = []
        return self._request_with_pydantic_model_response(
            method=HTTPMethod.POST,
            url=self.cfg.azul_url + "/api/v0/binaries",
            params=params,
            json={"hashes": hashes},
            response_model=models_restapi.EntityFind,
            get_data_only=True,
        )

    def find(
        self,
        term: str | None,
        *,
        max_entities: int | None = None,
        count_entities: bool | None = None,
        sort_prop: models_restapi.FindBinariesSortEnum | None = None,
        sort_asc: bool | None = None,
    ) -> models_restapi.EntityFind:
        """Find binaries matching the term query, limiting to the max_entities."""
        return self._base_find(
            term=term, max_entities=max_entities, count_entities=count_entities, sort_prop=sort_prop, sort_asc=sort_asc
        )

    def find_simple(
        self,
        find_options: FindOptions,
        *,
        max_entities: int | None = None,
        count_entities: bool | None = None,
        sort_prop: models_restapi.FindBinariesSortEnum | None = None,
        sort_asc: bool | None = None,
    ) -> models_restapi.EntityFind:
        """Find binaries matching the term query, limiting to the max_entities."""
        return self._base_find(
            term=find_options.to_query(),
            max_entities=max_entities,
            count_entities=count_entities,
            sort_prop=sort_prop,
            sort_asc=sort_asc,
        )

    def find_hashes(self, hashes: list[str]) -> models_restapi.EntityFind:
        """Check if a list of hashes are in Azul and returns their basic summary information."""
        return self._base_find(hashes=hashes)

    @dataclass
    class FindAll:
        """Result of a find_all query, with iterator to look at each binary."""

        approx_total: int = 0
        iter: Generator[models_restapi.EntityFindSimpleItem, None, None] = None

        def __iter__(self):
            """Iterator over the default iterator of iter."""
            return self.iter

    def find_all(
        self,
        find_options: FindOptions,
        *,
        max_binaries: int = 0,
        request_binaries: int = 5000,
    ) -> FindAll:
        """Find all binaries matching the term query, returned via a generator."""
        if max_binaries and max_binaries < request_binaries:
            request_binaries = max_binaries

        params = dict(
            term=find_options.to_query(),
            numodels_restapi=request_binaries,
        )
        params = self.filter_none_values(params)
        resp: models_restapi.EntityFindSimple = self._request_with_pydantic_model_response(
            method=HTTPMethod.POST,
            url=self.cfg.azul_url + "/api/v0/binaries/all",
            params=params,
            json={"after": None},
            response_model=models_restapi.EntityFindSimple,
            get_data_only=True,
        )

        def _iterate_binaries(params: dict, resp: models_restapi.EntityFindSimple):
            """Iterate over the found binaries."""
            after = resp.after
            found = 0
            while True:
                after = resp.after
                if len(resp.items) == 0 or not after:
                    return
                for item in resp.items:
                    yield item
                    found += 1
                    if max_binaries and found >= max_binaries:
                        # quit even if we have more than requested that we can supply
                        return
                resp: models_restapi.EntityFindSimple = self._request_with_pydantic_model_response(
                    method=HTTPMethod.POST,
                    url=self.cfg.azul_url + "/api/v0/binaries/all",
                    params=params,
                    json={"after": after},
                    response_model=models_restapi.EntityFindSimple,
                    get_data_only=True,
                )

        return BinariesMeta.FindAll(
            approx_total=resp.total if resp.total else 0,
            iter=_iterate_binaries(params, resp) if len(resp.items) > 0 else iter([]),
        )

    def get_model(self) -> models_restapi.EntityModel:
        """Return the model for the binaries."""
        return self._request_with_pydantic_model_response(
            method=HTTPMethod.GET,
            url=self.cfg.azul_url + "/api/v0/binaries/model",
            response_model=models_restapi.EntityModel,
            get_data_only=True,
        )

    def find_autocomplete(self, term: str) -> models_restapi.AutocompleteContext:
        """Looks for potential auto-completes for a search term and returns them.

        :param str term: Term to try and auto-complete.
        """
        # NOTE - offset is auto calculated as it's used for cursor position which makes no sense
        # from a client API perspective.

        return self._request_with_pydantic_model_response(
            method=HTTPMethod.GET,
            url=self.cfg.azul_url + "/api/v0/binaries/autocomplete",
            response_model=autocomplete_type_adapter,
            params={"term": term, "offset": len(term) - 1},
            get_data_only=True,
        )

    def get_has_newer_metadata(self, sha256: str, timestamp: str) -> models_restapi.BinaryDocuments:
        """Check if a binary has data newer than the provided timestamp in ISO8601 format."""
        return self._request_with_pydantic_model_response(
            method=HTTPMethod.GET,
            url=self.cfg.azul_url + f"/api/v0/binaries/{sha256}/new",
            response_model=models_restapi.BinaryDocuments,
            params={"timestamp": timestamp},
            get_data_only=True,
        )

    def get_similar_ssdeep_entities(self, ssdeep: str, *, max_matches: int = 20) -> models_restapi.SimilarFuzzyMatch:
        """Return id and similarity score of entities with a similar ssdeep fuzzyhash."""
        if not ssdeep:
            raise ValueError("ssdeep must be set to something.")
        return self._request_with_pydantic_model_response(
            method=HTTPMethod.GET,
            url=self.cfg.azul_url + "/api/v0/binaries/similar/ssdeep",
            response_model=models_restapi.SimilarFuzzyMatch,
            params={"ssdeep": ssdeep, "max_matches": max_matches},
            get_data_only=True,
        )

    def get_similar_tlsh_entities(self, tlsh: str, *, max_matches: int = 20) -> models_restapi.SimilarFuzzyMatch:
        """Return id and similarity score of entities with a similar tlsh fuzzyhash."""
        if not tlsh:
            raise ValueError("tlsh must be set to something.")
        return self._request_with_pydantic_model_response(
            method=HTTPMethod.GET,
            url=self.cfg.azul_url + "/api/v0/binaries/similar/tlsh",
            response_model=models_restapi.SimilarFuzzyMatch,
            params={"tlsh": tlsh, "max_matches": max_matches},
            get_data_only=True,
        )

    def get_similar_entities(self, sha256: str, *, recalculate: bool = False) -> models_restapi.SimilarMatch:
        """Return information about similar entities."""
        return self._request_with_pydantic_model_response(
            method=HTTPMethod.GET,
            url=self.cfg.azul_url + f"/api/v0/binaries/{sha256}/similar",
            response_model=models_restapi.SimilarMatch,
            params={"recalculate": recalculate},
            get_data_only=True,
        )

    def get_nearby_entities(
        self,
        sha256: str,
        *,
        include_cousins: models_restapi.IncludeCousinsEnum = models_restapi.IncludeCousinsEnum.Standard,
    ) -> models_restapi.ReadNearby:
        """Get information about nearby entities (used to build relational tree graph)."""
        return self._request_with_pydantic_model_response(
            method=HTTPMethod.GET,
            url=self.cfg.azul_url + f"/api/v0/binaries/{sha256}/nearby",
            response_model=models_restapi.ReadNearby,
            params={"include_cousins": include_cousins.value},
            get_data_only=True,
        )

    def get_binary_tags(self, sha256: str) -> models_restapi.ReadAllEntityTags:
        """Return all tags for an binary."""
        return self._request_with_pydantic_model_response(
            method=HTTPMethod.GET,
            url=self.cfg.azul_url + f"/api/v0/binaries/{sha256}/tags",
            response_model=models_restapi.ReadAllEntityTags,
            get_data_only=True,
        )

    def create_tag_on_binary(self, sha256: str, tag: str, security: str) -> None:
        """Attach a tag to the provided binaries sha256."""
        return self._request(
            method=HTTPMethod.POST,
            url=self.cfg.azul_url + f"/api/v0/binaries/{sha256}/tags/{tag}",
            json={"security": security},
        )

    def delete_tag_on_binary(self, sha256: str, tag: str) -> models_restapi.AnnotationUpdated:
        """Delete the specified tag from the specified binary."""
        return self._request_with_pydantic_model_response(
            method=HTTPMethod.DELETE,
            url=self.cfg.azul_url + f"/api/v0/binaries/{sha256}/tags/{tag}",
            response_model=models_restapi.AnnotationUpdated,
            get_data_only=True,
        )

    def get_binary_status(self, sha256: str) -> models_restapi.Status:
        """Get the plugin statuses for a binary."""
        return self._request_with_pydantic_model_response(
            method=HTTPMethod.GET,
            url=self.cfg.azul_url + f"/api/v0/binaries/{sha256}/statuses",
            response_model=models_restapi.Status,
            get_data_only=True,
        )

    def get_binary_documents(
        self, sha256: str, *, action: azm.BinaryAction | None = None, size: int = 1000
    ) -> models_restapi.OpensearchDocuments:
        """Get opensearch documents for a binary.

        :param str sha256: The sha256 of the binary to look for documents for.
        :param BinaryAction action: The action to get events for.
        :param int size: Maximum number of events that will be returned.
        """
        params = {"event_type": action, "size": size}
        params = self.filter_none_values(params)

        return self._request_with_pydantic_model_response(
            method=HTTPMethod.GET,
            url=self.cfg.azul_url + f"/api/v0/binaries/{sha256}/events",
            response_model=models_restapi.OpensearchDocuments,
            params=params,
            get_data_only=True,
        )



"""Mappings for the Azul user API."""

import logging
from http import HTTPMethod

from azul_bedrock import models_restapi
from azul_bedrock.models_auth import UserInfo

from azul_client.api.base_api import BaseApiHandler

logger = logging.getLogger(__name__)


class Users(BaseApiHandler):
    """API for accessing user information of Azul."""

    def get_opensearch_user_info(self) -> models_restapi.UserAccess:
        """Get Opensearch User info through Azul's restapi."""
        return self._request_with_pydantic_model_response(
            url=self.cfg.azul_url + "/api/v0/users/me/opensearch",
            method=HTTPMethod.GET,
            response_model=models_restapi.UserAccess,
        )

    def get_user_info(self) -> UserInfo:
        """Get user information for the current Token that Azul cares about."""
        return self._request_with_pydantic_model_response(
            url=self.cfg.azul_url + "/api/v0/users/me", method=HTTPMethod.GET, response_model=UserInfo
        )



"""A simple server that obtains the code generated by Keycloak.

A url is presented to the user, they navigate to it in a browser and proceed with authentication.
Then keycloak will redirect the browser to this web server and we extract the code generated by keycloak.
"""

import sys
import urllib.parse
from http.server import BaseHTTPRequestHandler, HTTPServer


class OIDCResponseServer(BaseHTTPRequestHandler):
    """Server to receive oidc callback to enable code exchange with PKCE."""

    expected_path = ""
    expected_state = ""
    token_code = None

    def do_GET(self):
        """Handle get request from users browser."""
        # expected like http://127.0.0.1:8080/authorisation-code/callback?code=XXXX&state=YYYY
        try:
            if not self.path.startswith(f"{OIDCResponseServer.expected_path}?"):
                raise Exception(f"bad path '{self.path}' not start with '{self.expected_path}?'")
            parsed = urllib.parse.urlparse(self.path)
            qs = urllib.parse.parse_qs(parsed.query)
            if OIDCResponseServer.expected_state != qs["state"][0]:
                raise Exception("Returned state does not match")
            OIDCResponseServer.token_code = qs["code"][0]

        except Exception as e:
            self.send_response(400)
            self.send_header("Content-type", "text/html")
            self.end_headers()
            self.wfile.write(bytes("<html><head><title>Azul Client Auth</title></head>", "utf-8"))
            self.wfile.write(bytes("<body>", "utf-8"))
            self.wfile.write(bytes("<p>Failure</p>", "utf-8"))
            self.wfile.write(bytes(f"<p>{str(e)}</p>", "utf-8"))
            self.wfile.write(bytes("</body></html>", "utf-8"))

        else:
            self.send_response(200)
            self.send_header("Content-type", "text/html")
            self.end_headers()
            self.wfile.write(bytes("<html><head><title>Azul Client Auth</title></head>", "utf-8"))
            self.wfile.write(bytes("<body>", "utf-8"))
            self.wfile.write(bytes("<p>Success</p>", "utf-8"))
            self.wfile.write(bytes("<p>You may now close this tab.</p>", "utf-8"))
            self.wfile.write(bytes("</body></html>", "utf-8"))


def receive_code(
    expected_state: str, *, path: str = "/authorisation-code/callback", hostname: str = "localhost", port: int = 8080
):
    """Receive one http request from users browser, parse code and shut down."""
    OIDCResponseServer.expected_path = path
    OIDCResponseServer.expected_state = expected_state
    server = HTTPServer((hostname, port), OIDCResponseServer)
    print("Server started at http://%s:%s. Waiting for oauth callback." % (hostname, port), file=sys.stderr)

    try:
        while not OIDCResponseServer.token_code:
            server.handle_request()
    except KeyboardInterrupt:
        pass
    server.server_close()
    print("Server stopped.")
    return OIDCResponseServer.token_code


if __name__ == "__main__":
    code = receive_code("51998", path="/test")
    print(f"{code=}")



from azul_bedrock import models_restapi

from azul_client.exceptions import BadResponse, BadResponse404

from .base_test import BaseApiTest


class TestPluginsApi(BaseApiTest):
    def setUp(self):
        super().setUp()

        # Ensure this is unset between runs
        self.api.set_excluded_security(None)

    @staticmethod
    def convert_plugins_to_dict(
        list_of_plugins: list[models_restapi.LatestPluginWithVersions],
    ) -> dict[str, models_restapi.PluginEntity]:
        """Helper functions to convert the list of plugins into a dictionary with plugin name as the key."""
        result = dict()
        for p in list_of_plugins:
            result[p.newest_version.name] = p.newest_version
        return result

    def test_get_all_plugins(self):
        """Verify that you can get a large list of the plugins installed and running on azul."""
        list_of_plugins = self.api.plugins.get_all_plugins()
        self.assertGreater(len(list_of_plugins), 10)
        self.assertGreaterEqual(len(list_of_plugins[0].versions), 1)
        self.assertIsInstance(list_of_plugins[0].newest_version, models_restapi.PluginEntity)

        dict_of_plugins = self.convert_plugins_to_dict(list_of_plugins)
        self.assertIn(self.entropy_plugin_name, dict_of_plugins.keys())
        self.assertIn(self.python_known_plugin_name, dict_of_plugins.keys())

    def test_security_exclusion_get_all_plugin(self):
        self.api.set_excluded_security([self.min_security])
        plugin_list_min = self.api.plugins.get_all_plugins()

        self.api.set_excluded_security([self.max_security])
        plugin_list_max = self.api.plugins.get_all_plugins()

        plugin_list_default = []
        try:
            self.api.set_excluded_security([self.default_security])
            plugin_list_default = self.api.plugins.get_all_plugins()
        except BadResponse404:
            pass

        # These should be large (all the official plugins minus maybe 1 or 2).
        self.assertGreater(len(plugin_list_min), 0)
        self.assertGreater(len(plugin_list_max), 0)
        # This may be 0 or a few
        self.assertGreaterEqual(len(plugin_list_default), 0)

        # when we cut out the default plugins that should remove a lot of plugins
        # So these numbers can't be equal and the default list may even be length 0.
        self.assertGreater(len(plugin_list_max), len(plugin_list_default))
        self.assertGreater(len(plugin_list_min), len(plugin_list_default))

    def test_get_all_plugin_statuses(self):
        """Test that you can get all the plugin statuses and that entropy has completed at least 100 jobs."""
        response_model = self.api.plugins.get_all_plugin_statuses()
        self.assertGreater(len(response_model), 10)

        checked_entropy = False
        for model in response_model:
            if model.newest_version.name == self.entropy_plugin_name:
                checked_entropy = True
                self.assertIsNotNone(model.last_completion)
                self.assertGreater(model.success_count, 100)

        self.assertTrue(checked_entropy, "Failed to find a status for entropy!")

    def test_get_plugin(self):
        """Verify that when you get plugins you can get their configurations or not if they don't have them."""
        list_of_plugins = self.api.plugins.get_all_plugins()
        dict_of_plugins = self.convert_plugins_to_dict(list_of_plugins)
        latest_version = dict_of_plugins[self.entropy_plugin_name].version
        plugin_info = self.api.plugins.get_plugin(self.entropy_plugin_name, latest_version)
        self.assertGreater(plugin_info.num_entities, 1)
        self.assertIsInstance(plugin_info.plugin.config, dict)
        self.assertGreaterEqual(len(plugin_info.status), 1)

        python_latest_version = dict_of_plugins[self.python_known_plugin_name].version
        plugin_info = self.api.plugins.get_plugin(self.python_known_plugin_name, python_latest_version)
        self.assertGreater(plugin_info.num_entities, 1)
        self.assertIsInstance(plugin_info.plugin.config, dict)
        self.assertGreaterEqual(len(plugin_info.status), 1)



"""Config handling."""

import configparser
import json
import os
import sys
import tempfile

import click
import pydantic
from filelock import FileLock
from pydantic_settings import BaseSettings

config_section = "default"


def switch_section(section: str):
    """Switch to a different azul deployment configured in the ini."""
    global config_section
    config_section = section


@click.group(name="config")
def _client_config():
    """Change azul-client configuration."""
    pass


class ConfigLocation(BaseSettings):
    """Path to settings and lock files for azul."""

    azul_config_location: str = os.path.join(os.path.expanduser("~"), ".azul.ini")
    token_refresh_path_lock: str = os.path.join(tempfile.gettempdir(), "azul-token-refresh.lock")
    token_lock_timeout: float = 30


config_location = ConfigLocation()
# NOTE - double locking with this as a decorator does not harm because it tracks the current processes PID.
# So if it double acquires the lock it knows it already has the lock and continues to work.
_lock_azul_config = FileLock(config_location.token_refresh_path_lock, timeout=config_location.token_lock_timeout)


class Config(BaseSettings):
    """Config wrapper."""

    azul_url: str = "http://localhost"
    oidc_url: str = "http://keycloak/.well-known/openid-configuration"
    auth_type: str = "callback"
    auth_scopes: str = ""
    auth_client_id: str = "azul-web"
    auth_client_secret: str = ""  # noqa S105
    azul_verify_ssl: bool = True
    auth_token: dict | None = {}
    auth_token_time: int = 0
    max_timeout: float = 300.0
    oidc_timeout: float = 10.0

    @pydantic.field_validator("azul_url")
    def no_trailing_slash(cls, v):
        """Remove trailing slash from azul_url."""
        return v.rstrip("/")

    @_lock_azul_config
    def save(self):
        """Save the current configuration."""
        tmp = self.model_dump()
        # save auth token as json string
        tmp["auth_token"] = json.dumps(tmp["auth_token"])

        location = ConfigLocation().azul_config_location
        cfg = configparser.ConfigParser()
        cfg.read(location)
        cfg[config_section] = tmp
        with open(ConfigLocation().azul_config_location, "w") as configfile:
            cfg.write(configfile)


@_client_config.command()
@_lock_azul_config
def clear_auth():
    """Reset current auth information."""
    conf = get_config()
    conf.auth_token = {}  # noqa S105
    conf.auth_token_time = 0
    conf.save()


@_lock_azul_config
def get_config():
    """Get config loaded from file."""
    location = ConfigLocation().azul_config_location
    if not os.path.exists(location):
        print(f"ERROR - no config found - generating default at {location}", file=sys.stderr)
        print("You will likely need to edit this config.", file=sys.stderr)
        conf = Config()
        conf.save()

    print(f"Loading config [{config_section}] from {location}", file=sys.stderr)
    tmp = configparser.ConfigParser()
    tmp.read(location)
    conf = {}
    try:
        # configparser has an odd data structure, convert to dictionary
        conf = {**tmp[config_section]}
    except KeyError as e:
        if config_section == "default":
            print(f"Config section [{config_section}] is invalid, generating defaults", file=sys.stderr)
        else:
            raise Exception(f"config section [{config_section}] is invalid") from e

    # the auth token was saved as a json string
    if conf.get("auth_token"):
        conf["auth_token"] = json.loads(conf["auth_token"])
    config = Config(**conf)
    print(f"using Azul API at {config.azul_url}\n", file=sys.stderr)
    return config



"""Mappings for the Azul purge API."""

import logging
from http import HTTPMethod

import pendulum
from azul_bedrock import models_restapi
from pendulum.parsing.exceptions import ParserError

from azul_client.api.base_api import BaseApiHandler

logger = logging.getLogger(__name__)


class Purge(BaseApiHandler):
    """API for purging files out of Azul, this API requires a user who is an admin to use it.

    NOTE - Purges should not be scripted as standard practice, they have a high cost to Azul
    and too many purges or frequent re-purging will cause instability in Azul.
    """

    def purge_submission(
        self, track_source_references: str, *, timestamp: str, purge: bool = False
    ) -> models_restapi.PurgeSimulation | models_restapi.PurgeResults:
        """Purge a set of submissions (requires admin privileges).

        :param str source_data_kvs: Unique identifier for the submission set (submission source + all metadata).
        :param str binary: Sha256 Id of the binary to purge from the submission (optional).
        :param str timestamp: Timestamp of the submission to purge (optional).
        :param bool purge: If true, perform the purge instead of a simulation (default False).
        """
        params = {"purge": purge}
        if not timestamp:
            raise ValueError("Timestamp is required to be set and cannot be None or and empty string.")
        try:
            pendulum.parse(timestamp)
        except ParserError as e:
            raise ValueError(f"Timestamp has an invalid value '{timestamp}' must be a") from e
        params["timestamp"] = timestamp

        if purge:
            response_model = models_restapi.PurgeResults
        else:
            response_model = models_restapi.PurgeSimulation

        return self._request_with_pydantic_model_response(
            url=self.cfg.azul_url + f"/api/v0/purge/submission/{track_source_references}",
            method=HTTPMethod.DELETE,
            response_model=response_model,
            params=params,
            get_data_only=True,
        )

    def purge_link(
        self, track_link: str, *, purge: bool = False
    ) -> models_restapi.PurgeSimulation | models_restapi.PurgeResults:
        """Purge a manually added relationship between binaries."""
        params = {"purge": purge}

        if purge:
            response_model = models_restapi.PurgeResults
        else:
            response_model = models_restapi.PurgeSimulation

        return self._request_with_pydantic_model_response(
            url=self.cfg.azul_url + f"/api/v0/purge/link/{track_link}",
            method=HTTPMethod.DELETE,
            response_model=response_model,
            params=params,
            get_data_only=True,
        )



"""Mappings for the Azul plugins API."""

import logging
from http import HTTPMethod

from azul_bedrock import models_restapi
from pydantic import TypeAdapter

from azul_client.api.base_api import BaseApiHandler

logger = logging.getLogger(__name__)

latest_plugin_list_adapter = TypeAdapter(list[models_restapi.LatestPluginWithVersions])
latest_plugin_status_list_adapter = TypeAdapter(list[models_restapi.PluginStatusSummary])


class Plugins(BaseApiHandler):
    """API for listing out the plugins currently running in Azul, their configuration and latest statuses."""

    def get_all_plugins(self) -> list[models_restapi.LatestPluginWithVersions]:
        """Read names and versions of all registered plugins."""
        return self._request_with_pydantic_model_response(
            url=self.cfg.azul_url + "/api/v0/plugins",
            method=HTTPMethod.GET,
            response_model=latest_plugin_list_adapter,
            get_data_only=True,
        )

    def get_all_plugin_statuses(self) -> list[models_restapi.PluginStatusSummary]:
        """Read names and versions of all registered plugins and count statuses.

        Note - the status count is inaccurate because it doesn't filter out duplicates.
        A duplicate is where the same binary is submitted to a plugin with a different path.
        """
        return self._request_with_pydantic_model_response(
            url=self.cfg.azul_url + "/api/v0/plugins/status",
            method=HTTPMethod.GET,
            response_model=latest_plugin_status_list_adapter,
            get_data_only=True,
        )

    def get_plugin(self, name: str, version: str) -> models_restapi.PluginInfo:
        """Get all the configuration information about a specific plugin by name and version."""
        return self._request_with_pydantic_model_response(
            url=self.cfg.azul_url + f"/api/v0/plugins/{name}/versions/{version}",
            method=HTTPMethod.GET,
            response_model=models_restapi.PluginInfo,
            get_data_only=True,
        )



"""Mappings for the Azul sources API."""

import logging
from http import HTTPMethod

from azul_bedrock import models_restapi
from azul_bedrock import models_settings as azs
from pydantic import TypeAdapter

from azul_client.api.base_api import BaseApiHandler

logger = logging.getLogger(__name__)


source_dict_type_adapter = TypeAdapter(dict[str, azs.Source])


class Sources(BaseApiHandler):
    """API for accessing information about the sources configured in Azul."""

    def get_all_sources(self) -> dict[str, azs.Source]:
        """Get a list of all sources from Azul."""
        return self._request_with_pydantic_model_response(
            url=self.cfg.azul_url + "/api/v0/sources",
            method=HTTPMethod.GET,
            response_model=source_dict_type_adapter,
            get_data_only=True,
        )

    def check_source_exists(self, source_id: str) -> bool:
        """Check if the provided source exists in Azul."""
        return self._generic_head_request(self.cfg.azul_url + f"/api/v0/sources/{source_id}")

    def read_source(self, source_id: str) -> models_restapi.Source:
        """Read the information for the requested source."""
        return self._request_with_pydantic_model_response(
            url=self.cfg.azul_url + f"/api/v0/sources/{source_id}",
            method=HTTPMethod.GET,
            response_model=models_restapi.Source,
            get_data_only=True,
        )

    def read_source_references(self, source_id: str, *, term: str = "") -> models_restapi.References:
        """Read source references for a source_id with an optional term query."""
        return self._request_with_pydantic_model_response(
            url=self.cfg.azul_url + f"/api/v0/sources/{source_id}/references",
            method=HTTPMethod.GET,
            response_model=models_restapi.References,
            params={"term": term},
            get_data_only=True,
        )



import multiprocessing
import unittest
from concurrent.futures import ThreadPoolExecutor

from azul_bedrock import models_restapi

from azul_client import Api
from azul_client.config import _lock_azul_config, get_config

from . import module_ref, module_sha256s, module_source


def setup_api_and_upload_file(content: bytes) -> models_restapi.BinaryData:
    """Create the azul api and upload to Azul.

    This is done so all multiprocesses are interacting with the .azul.ini simultaneously."""
    api = Api(get_config())
    upload_kwargs = {
        "source_id": module_source,
        "filename": "test_multiprocessing_upload",
        "security": "OFFICIAL",
        "references": module_ref,
    }
    resp = api.binaries_data.upload(content, **upload_kwargs)
    print(f"uploaded {content.decode()[:4]}")
    return resp


@_lock_azul_config
def update_azul_ini():
    conf = get_config()
    conf.auth_token_time = 0
    conf.save()


def save_to_azul_init(_index):
    """Method to attempt to modify and update .azul.ini to try and compete with other processes."""
    try:
        for _i in range(4):
            update_azul_ini()
        return "success"
    except Exception as e:
        print(f"Failed to update azul init with error {e}")
        return "fail"


class MultiprocessingAndThreadingUploadTest(unittest.TestCase):
    def test_multiprocessing_upload(self):
        """Verify that multiple files can be simultaneously uploaded to Azul with the azul client."""
        total_files_to_upload = 20
        input_content_list = []
        for input_content in range(total_files_to_upload):
            input_content_list.append(
                (
                    f"{input_content} - Basic and very small amount of content for input"
                    + f" to upload in a multiprocessing context {input_content}" * 20
                ).encode()
            )
        # Perform test
        m_ctx = multiprocessing.get_context("forkserver")
        with m_ctx.Pool(processes=4) as pool:
            output_results = pool.map(setup_api_and_upload_file, input_content_list)

            self.assertEqual(len(output_results), total_files_to_upload)
            # All sha256's should be uploaded.
            output_sha256s = [r.sha256 for r in output_results]
            print(output_sha256s)
            self.assertCountEqual(
                output_sha256s,
                [
                    "325f80b4bd8977075d0581df5d7f2a12d2271487be7d3da2d94536aa615e5aff",
                    "c42840a3fc99138aca7136de0378649c8035a2331dfdb00676c5e76fbddba3f3",
                    "238e269ad74b1b8fd15b4340810029d5f1d054bbef28108097303c9e99b77337",
                    "2bf41b76afab954e4f0569fc0fc223141057eb35e1ed72e29e400d140eb3e1fb",
                    "bf06cde1e4a1c1e4e3ab67c8d99786d5f4784a7b94d5a59a458faac945a4d62d",
                    "382f80b87604dc0eda0052c6a7c5499b93db43af96d27904ed6e0313efa21646",
                    "b5a85307882cbf990760b687841f2138a0eb302e6dbf96ac47a97e351b6bf14c",
                    "8196803e2bdb5fac475a22b592e376dbb71c27ff14cb69d1007378362c1c068b",
                    "bb54f326d59198c7ad41527d4f9754a58f1c3c5fb0b428f740de5e4fc599b7e8",
                    "a0bd01e1a73aa786e2bab6f6022ad72d3bf7e29c87a3879b9a85b2923cd06eec",
                    "247428159ec663342f02d04d8e9792566d16ac013aa3ae88c0677a6fbe52f188",
                    "33915c3b529c8b33495319ea60b1b09e5a331ac0f56721c3e5defb713f6712ff",
                    "ef8b6225b831cf09a736000e6468fdd2cde3c0680481453463735e757755e585",
                    "49e60ab6a45580a94723e05466cb1f89de821b5aa2c99f9e6cb147b44caf03a1",
                    "521f9c8d411ba075927fa40422c177668bbc407d3f15979df244e1230d55e70c",
                    "b10c25c5e8d307df2135f2e9f3794443e3c604f71b26560302c242cb380b24ab",
                    "5439f697ccb7fbb15ba8d2f3720f5ebdc1788a14c449ff39fe8dfaf7a9ffb15a",
                    "a1148fac415968baad75b840a433ad06ed5d27eea9bc4697560adf26667475e1",
                    "e5eb2757cb6f6eab813803f1a72aca0790086702785b86c221247a9365337500",
                    "9097aa562ffa73ce30353cf6f2e40801ac3f4b084750a497e8b6835f7d6693a6",
                ],
            )
            # Verify file formats are all equal
            file_format = list(set(r.file_format for r in output_results))
            self.assertEqual(len(file_format), 1)
            self.assertEqual(file_format[0], "text/plain")
            # Verify mimes are all equal
            file_mime = list(set(r.mime for r in output_results))
            self.assertEqual(len(file_mime), 1)
            self.assertEqual(file_mime[0], "text/plain")

    def test_multiprocessing_modify_azul_ini(self):
        """Have lots of processes simply try and modify .azul.ini and nothing breaks."""
        total_modifications = 20
        m_ctx = multiprocessing.get_context("forkserver")
        with m_ctx.Pool(processes=4) as pool:
            output_results = pool.map(save_to_azul_init, range(total_modifications))
            result = list(output_results)
            self.assertEqual(len(result), total_modifications)
            self.assertEqual(len(set(result)), 1)
            self.assertEqual(result[0], "success")

    def test_threading_upload(self):
        """Verify that multiple files can be simultaneously uploaded to Azul with the azul client."""
        total_files_to_upload = 20
        input_content_list = []
        for input_content in range(total_files_to_upload):
            input_content_list.append(
                (
                    f"{input_content} - Basic and very small amount of content for input"
                    + f" to upload in a threading context {input_content}" * 20
                ).encode()
            )
        # Perform test
        with ThreadPoolExecutor(max_workers=4) as executor:
            output_results = executor.map(setup_api_and_upload_file, input_content_list)
            output_results = list(output_results)

            self.assertEqual(len(output_results), total_files_to_upload)
            # All sha256's should be uploaded.
            output_sha256s = [r.sha256 for r in output_results]
            print(output_sha256s)
            self.assertCountEqual(
                output_sha256s,
                [
                    "a9e59630b94d384de85f38049ebd9fc98c5dc28d0a9a64c9021398d294e3f71d",
                    "1566c0d8e42cf66263fb4241a4d6818afaad72ad5c207d3a5cded4c97e67199e",
                    "3aca772aced98400c1dbda55150bff9a5524a9f2214889e41c75b1e18d0ef710",
                    "f479f8735372dddb2d11a390fc9b871c666e658664641b4991090d6b1f987544",
                    "2a82005c9c66a05ab00e456313381ff25aff0d12649d67b62ae8b8d35f981ad9",
                    "2cb79026088e754eb1f75d67c97ef0199751994ef824c1cea47c72c21af8e77d",
                    "2a56f5d82227d7a8e5bc87c478700315263f933d8a11de8c76ed96852579d914",
                    "cdbfb8e59ba8c65a3918af46f5c4ad0ca6dd161ddfe39a0de3037a7bf18f13e8",
                    "023d17f35aab8284c7735532326d47236c75b602850732edf4f098dc2f008943",
                    "f443f56d63728e058721a5731ce4d0ce2f4603a5bf33f55fb2586bd2b23a026b",
                    "c1134e078894141c1e9e1cf5d223c57bc281453a30d5962c1420cf1b2726f2ce",
                    "76289cd60240c81a9f3408513c514c59b40be8ad8de86462da4563c95406ec49",
                    "4f41d51506d3918ea6c9f537c4cd7f73a356a0b84b5537d71ba9882e06ca6dc2",
                    "89a9fa723a6eb00d897c5a867278092bab39d2f8300fe2a14c4ea06bfe958a9c",
                    "c3802bb20aedc1ad3ba9cabb8cdc9dbb33298f7512348bd36a606c6d9eb97c23",
                    "385b5a2b22c5e721fecf5c26f77ef6ef05d6eccf035e3b022fa2905a107cc001",
                    "b4166b23982b2c41455ddd68058480dc7e9c7ab983c674d4f2630ad99a9df60b",
                    "b5a62ab875dce14c97d4862d78fe35e84ab61cb95332f37d430e5b4f4bdd0ae9",
                    "5471b7bc733de74b73abdc80440307696ea9118ddac772d5fd0f70a35f5b67e1",
                    "5443f2194d2a2d2010022571ffa02a108818d25286efd86f892f56c9dcdf719a",
                ],
            )
            # Verify file formats are all equal
            file_format = list(set(r.file_format for r in output_results))
            self.assertEqual(len(file_format), 1)
            self.assertEqual(file_format[0], "text/plain")
            # Verify mimes are all equal
            file_mime = list(set(r.mime for r in output_results))
            self.assertEqual(len(file_mime), 1)
            self.assertEqual(file_mime[0], "text/plain")

    def test_threading_modify_azul_ini(self):
        """Have lots of threads simply try and modify .azul.ini and nothing breaks."""
        total_modifications = 20
        with ThreadPoolExecutor(max_workers=4) as executor:
            output_results = executor.map(save_to_azul_init, range(total_modifications))
            result = list(output_results)
            self.assertEqual(len(result), total_modifications)
            self.assertEqual(len(set(result)), 1)
            self.assertEqual(result[0], "success")



"""Auth handling."""

import base64
import hashlib
import json
import os
import random
import re
import sys
import time
import urllib.parse

import httpx

from azul_client import config

from . import callback


def _get_json(resp: httpx.Response, dbg: str) -> dict:
    """Try to read json data out of the web request."""
    if resp.status_code != 200:
        raise Exception(f"Failed to fetch {dbg} - Code: {resp.status_code} - Content:\n{resp.content[:1000]}")
    if "application/json" not in resp.headers.get("content-type"):
        raise Exception(
            f"Failed to fetch {dbg} - response mime is not json ({resp.headers.get('content-type')}). "
            f"Is '{resp.url}' the correct endpoint? - "
            f"Content:\n{resp.content[:1000]}"
        )
    try:
        data = resp.json()
    except json.decoder.JSONDecodeError as e:
        raise Exception(
            f"Failed to fetch {dbg} - response was not json. "
            f"Is '{resp.url}' the correct endpoint? - "
            f"Content:\n{resp.content[:1000]}"
        ) from e
    return data


class OIDC:
    """Handle authentication with Azul api."""

    def __init__(self, cfg: config.Config) -> None:
        self.cfg = cfg
        self._oidc_info = None
        # Requests used to force retries on status codes [500, 502, 503, 504]
        verify = True
        if not self.cfg.azul_verify_ssl:
            print("NO VERIFY SSL", file=sys.stderr)
            verify = False
        # Setup client with retries.
        self._verify = verify
        self._local_client = httpx.Client(
            timeout=httpx.Timeout(timeout=self.cfg.max_timeout),
        )

    def _get_oidc_info(self):
        """Return oidc json document containing locations of required resources for auth."""
        if not self._oidc_info:
            self._fetch_well_known()
        return self._oidc_info

    def _get_authorization_endpoint(self):
        return self._get_oidc_info()["authorization_endpoint"]

    def _get_token_endpoint(self):
        return self._get_oidc_info()["token_endpoint"]

    def _fetch_well_known(self):
        resp = self._local_client.get(self.cfg.oidc_url, timeout=httpx.Timeout(timeout=self.cfg.oidc_timeout))
        self._oidc_info = _get_json(resp, "well known")

    @config._lock_azul_config
    def get_client(self):
        """Return a httpx client object with an up to date authorization token."""
        out_client = httpx.Client(
            headers={"authorization": "Bearer " + self._get_access_token()},
            mounts={
                "http://": httpx.HTTPTransport(retries=5),
                "https://": httpx.HTTPTransport(retries=5),
            },
            timeout=httpx.Timeout(timeout=self.cfg.max_timeout),
            verify=self._verify,
        )
        return out_client

    def _via_service_token(self):
        """Retrieve a token from OIDC provider using service account flow."""
        resp = self._local_client.post(
            self._get_token_endpoint(),
            data={
                "response_type": "token",
                "client_id": self.cfg.auth_client_id,
                "client_secret": os.environ.get("AZUL_CLIENT_SECRET") or self.cfg.auth_client_secret,
                "grant_type": "client_credentials",
                "scope": self.cfg.auth_scopes,
            },
            timeout=httpx.Timeout(timeout=self.cfg.oidc_timeout),
        )
        return _get_json(resp, "via service token")

    def _via_code_callback(self):
        """Retrieve a token from OIDC provider using code callback."""
        # prove that requestor of token is same as receiver with code challenge and verify
        code_verifier = base64.urlsafe_b64encode(os.urandom(40)).decode("utf-8")
        code_verifier = re.sub("[^a-zA-Z0-9]+", "", code_verifier)

        code_challenge = hashlib.sha256(code_verifier.encode("utf-8")).digest()
        code_challenge = base64.urlsafe_b64encode(code_challenge).decode("utf-8")
        code_challenge = code_challenge.replace("=", "")

        port = 8080
        callback_url = f"http://localhost:{port}/client/callback"

        state = str(random.randint(1000000, 9999999))  # noqa S311
        params = {
            "response_type": "code",
            "client_id": self.cfg.auth_client_id,
            "redirect_uri": callback_url,
            "state": state,
            "scope": self.cfg.auth_scopes,
            "code_challenge": code_challenge,
            "code_challenge_method": "S256",
        }
        url_ask_user = self._get_authorization_endpoint() + "?" + urllib.parse.urlencode(params)
        print(f"Please navigate to the following url to continue authentication:\n{url_ask_user}", file=sys.stderr)
        code = callback.receive_code(state, path="/client/callback", hostname="localhost", port=port)
        if not code:
            raise Exception("No token retrieval code was returned")

        resp = self._local_client.post(
            self._get_token_endpoint(),
            data={
                "grant_type": "authorization_code",
                "code": code,
                "client_id": self.cfg.auth_client_id,
                "state": state,
                "scope": self.cfg.auth_scopes,
                "redirect_uri": callback_url,
                "code_verifier": code_verifier,
            },
            timeout=httpx.Timeout(timeout=self.cfg.oidc_timeout),
        )
        return _get_json(resp, "via code callback")

    def _via_refresh(self, tk: dict):
        """Obtain new tokens using previously issued refresh token."""
        refresh_token = tk.get("refresh_token", None)
        # Full re-auth if there is no refresh token
        if not refresh_token:
            return self._get_token_non_refresh()
        # need to refresh the current token
        resp = self._local_client.post(
            self._get_token_endpoint(),
            data={
                "grant_type": "refresh_token",
                "client_id": self.cfg.auth_client_id,
                "refresh_token": refresh_token,
                "scope": self.cfg.auth_scopes,
            },
            timeout=httpx.Timeout(timeout=self.cfg.oidc_timeout),
        )
        if 400 <= resp.status_code < 500:
            # maybe refresh token has expired, retry full auth
            return None
        return _get_json(resp, "via refresh")

    def _get_token_non_refresh(self):
        # use auth method nominated in config
        atype = self.cfg.auth_type
        if atype == "none":
            # no access token is required
            tk = {}
        elif atype == "callback":
            tk = self._via_code_callback()
        elif atype == "service":
            tk = self._via_service_token()
        else:
            raise NotImplementedError(atype)
        return tk

    def _get_token(self):
        """Get auth tokens to Azul."""
        if self.cfg.auth_type == "none":
            # no access token is required
            return {}

        if self.cfg.auth_token:
            if time.time() > self.cfg.auth_token_time + 60:
                # refresh the token
                tk = self._via_refresh(self.cfg.auth_token)
                if not tk:
                    # maybe refresh token has expired
                    print("Warning - Refresh token likely has expired.", file=sys.stderr)
                    tk = self._get_token_non_refresh()

            elif time.time() < self.cfg.auth_token_time + 60:
                # return current token
                return self.cfg.auth_token
        else:
            tk = self._get_token_non_refresh()

        self.cfg.auth_token = tk
        self.cfg.auth_token_time = int(time.time())
        # save updated token to config file
        self.cfg.save()
        return tk

    def _get_access_token(self) -> str:
        """Get valid access token to authenticate with Azul."""
        token = self._get_token().get("access_token", "none")
        if not token and self.cfg.auth_type != "none":
            raise Exception("Token required but was not found")
        return token



"""High level library flow."""

import json
import os
import re
import sys
from tempfile import SpooledTemporaryFile

import click
import pendulum
from azul_bedrock import models_restapi
from pendulum.datetime import DateTime
from pydantic import BaseModel
from rich.console import Console

from azul_client import config
from azul_client.api import Api
from azul_client.config import _client_config
from azul_client.exceptions import BadResponse, BadResponse404

api: Api = None

SECURITY_STRING_DESCRIPTION = "simple security string (use `azul security` to see available security strings)"
TIMESTAMP_DESCRIPTION = (
    "timestamp for which the file being submitted was sourced in ISO8601 format e.g 2025-05-26T02:11:44Z"
)


@click.group()
@click.option("-c", default="default", help="switch to a different configured Azul instance.")
def cli(c: str):
    """Interact with the Azul API via CLI tools."""
    global api
    config.switch_section(c)
    api = Api()


@click.group()
def binaries():
    """Upload, download and get metadata associated with binaries."""
    pass


@click.command(name="security")
@click.option("--full", is_flag=True, show_default=True, default=False, help="show full configuration")
def security(full: bool):
    """List Azul security classification settings."""
    settings = api.security.get_security_settings()
    if not full and settings.get("presets"):
        click.echo("Security Presets:")
        click.echo("\n".join(settings.get("presets")))
    else:
        click.echo(json.dumps(settings, indent=2))


@click.group(name="sources")
def sources():
    """List and get information about specific sources."""
    pass


@sources.command(name="list")
def sources_list():
    """List all of the source ids."""
    all_sources = api.sources.get_all_sources()
    click.echo("Source IDS:")
    click.echo("\n".join(all_sources.keys()))


@sources.command(name="full")
def sources_full():
    """Get the full source information for each source."""
    all_sources = api.sources.get_all_sources()
    all_sources_dumped = {}
    for k, val in all_sources.items():
        all_sources_dumped[k] = val.model_dump()
    click.echo("Sources:")
    click.echo(json.dumps(all_sources_dumped, indent=2))


@sources.command(name="info")
@click.argument("source")
def sources_info(source: str):
    """Get summary information about a specific SOURCE by source Id."""
    all_sources = api.sources.get_all_sources()
    for source_id, sourceObj in all_sources.items():
        if source_id.lower() == source.lower():
            click.echo(source_id + ":")
            click.echo("Description: " + sourceObj.description)
            click.echo("Submissions Expire After " + sourceObj.expire_events_after)
            click.echo("References:")
            for ref in sourceObj.references:
                click.echo(f"  name: '{ref.name}'")
                click.echo(f"  description: '{ref.description}'")
                click.echo(f"  required: '{ref.required}'")
            break


@click.group()
def plugins():
    """List and get information for plugins in Azul."""
    pass


@plugins.command(name="list")
def plugins_list():
    """List all of the plugins registered in Azul."""
    plugin_list = api.plugins.get_all_plugins()
    click.echo("Plugins (name version):")
    for p in plugin_list:
        click.echo(f"{p.newest_version.name} {p.newest_version.version}")


@plugins.command(name="info")
@click.argument("name")
@click.option("--version", type=str, help="version of the plugin to get info for (defaults to newest)")
def plugin_info(name: str, version: str):
    """Get the details of a plugin with the provided plugin name."""
    if version:
        try:
            details = api.plugins.get_plugin(name, version)
        except BadResponse404:
            click.echo(f"Plugin {name} {version} does not exist check the version and name.")
            return
        except BadResponse as e:
            click.echo(f"Plugin {name} could not be found due to error {e.message}.")
        click.echo(f"Providing detail for plugin {name} {version}")
        click.echo(details.plugin.model_dump_json(indent=2))

    try:
        plugin_list = api.plugins.get_all_plugins()
    except BadResponse as e:
        click.echo(f"Plugin {name} could not be found due to error {e.message}.")
        return

    for p in plugin_list:
        if p.newest_version.name == name:
            click.echo(f"Providing detail for plugin {p.newest_version.name} {p.newest_version.version}")
            click.echo(p.newest_version.model_dump_json(indent=2))
            return
    click.echo(f"Plugin {name} could not be found, check the name is valid.")


@binaries.command()
@click.argument("sha256")
def check(sha256: str):
    """Check if binary metadata associated with the provided SHA256 is in Azul or not."""
    if api.binaries_meta.check_meta(sha256):
        click.echo("Binary metadata available")
    else:
        click.echo("Binary metadata NOT available")
        sys.exit(1)


@binaries.command()
@click.argument("sha256")
def check_data(sha256: str):
    """Check if a binary in Azul has the original file stored in Azul for the provided SHA256."""
    if api.binaries_data.check_data(sha256):
        click.echo("Binary data available")
    else:
        click.echo("Binary data NOT available")
        sys.exit(1)


def _walk_files_in_path(path: str) -> list[str]:
    """Walks a user given path for files."""
    input_files = []
    if os.path.isdir(path):
        for root, dirs, files in os.walk(path, followlinks=False):
            files = [f for f in files if not f[0] == "."]
            dirs[:] = [d for d in dirs if not d[0] == "."]
            for name in files:
                loc = os.path.join(root, name)
                input_files.append(loc)
    elif os.path.isfile(path):
        input_files.append(path)
    else:
        raise Exception("cannot upload something that is not a folder or file")

    return input_files


def _shared_submit(
    confirmed: bool,
    path: str,
    *,
    security: str = "",
    timestamp: str = "",
    extract: bool = False,
    extract_password: str = "",
    parent: str = "",
    parent_rels: dict = None,
    source: str = "",
    source_refs: dict = None,
):
    """Common class for submitting binaries to Azul."""
    security = security if security else ""
    if not timestamp:
        timestamp = pendulum.now(pendulum.UTC).to_iso8601_string()
    else:
        timestamp = pendulum.parse(timestamp).to_iso8601_string()

    raw_input_files = _walk_files_in_path(path)

    # generate azul file names
    input_files = []
    for filepath in raw_input_files:
        # try to remove provided path, unless that was a reference to a specific file
        # in which case keep the filename only
        adjusted = filepath.removeprefix(path)
        filename = os.path.basename(filepath)
        if filename not in adjusted:
            adjusted = filename
        input_files.append((filepath, adjusted))

    # print info and confirm to upload
    click.echo(f"{len(input_files)} files found including:")
    for _, filepath in input_files[:10]:
        click.echo(filepath)

    click.echo(f"Security: {security}")
    click.echo(f"Timestamp: {timestamp}")
    click.echo(f"Extract: {extract}")
    click.echo(f"Extract Password: {extract_password}")
    if parent:
        click.echo(f"Parent: {parent}")
        click.echo(f"Relationship: {parent_rels}")
    else:
        click.echo(f"Source: {source}")
        click.echo(f"References: {source_refs}")

    if not confirmed and not click.confirm(f"Proceed with upload of {len(input_files)} files?"):
        sys.exit(1)

    # submit each file
    for fullpath, filepath in input_files:
        with open(fullpath, "rb") as f:
            if parent:
                resp = api.binaries_data.upload_child(
                    f,
                    parent_sha256=parent,
                    relationship=parent_rels,
                    security=security,
                    filename=filepath,
                    timestamp=timestamp,
                    extract=extract,
                    password=extract_password,
                )
            else:
                resp = api.binaries_data.upload(
                    f,
                    security=security,
                    source_id=source,
                    filename=filepath,
                    timestamp=timestamp,
                    references=source_refs,
                    extract=extract,
                    password=extract_password,
                )
            click.echo(f"{filepath} - {resp.sha256}")


def _print_model(model: BaseModel, pretty: bool):
    """Prints a Pydantic model for user consumption, with a pretty filter as required."""
    if pretty:
        # Configure our environment for using a pager if possible
        if "MANPAGER" not in os.environ and "PAGER" not in os.environ:
            os.environ["PAGER"] = "less -r"

        # Guess to see if the pager we are using supports color
        colour_supported = "less -r" in os.environ.get("MANPAGER", "") or "less -r" in os.environ.get("PAGER", "")

        console = Console()
        # Enable a pager for the entity document (its big) if this is an interactive terminal
        if console.is_terminal:
            with console.pager(styles=colour_supported):
                console.print(model)
        else:
            console.print(model)
    else:
        # Dump the entire JSON document for use with e.g. jq
        click.echo(model.model_dump_json(indent=4))


@binaries.command()
@click.argument("path")
@click.option("-y", is_flag=True, show_default=True, default=False, help="no confirmation prompt")
@click.option("--timestamp", type=str, help=TIMESTAMP_DESCRIPTION)
@click.option("--security", required=True, type=str, help=SECURITY_STRING_DESCRIPTION)
@click.option("--parent", required=True, type=str, help="SHA256 of parent file")
@click.option(
    "-r",
    "--relationship",
    required=True,
    multiple=True,
    type=str,
    help="""relationship information between the uploaded child and the parent in form key:value e.g:
    azul put-child --relationship action:extracted --relationship relationship:friend
    """,
)
@click.option(
    "--extract",
    is_flag=True,
    show_default=True,
    default=False,
    help="extract the provided child file (must be trusted archive)",
)
@click.option("--extract-password", type=str, help="password to use when extracting the child archive")
def put_child(
    y: bool,
    path: str,
    timestamp: str,
    security: str,
    parent: str,
    relationship: list[str],
    extract: bool,
    extract_password: str,
):
    """Uploads a binary from PATH as a child of a pre-existing parent binary."""
    parsed_relationships = [r.split(":", 1) for r in relationship]
    relation_dict = {item[0]: item[1] for item in parsed_relationships}
    _shared_submit(
        y,
        path,
        security=security,
        timestamp=timestamp,
        parent=parent,
        parent_rels=relation_dict,
        extract=extract,
        extract_password=extract_password,
    )


@binaries.command()
@click.option("-y", is_flag=True, show_default=True, default=False, help="no confirmation prompt")
@click.argument("path")
@click.argument("source")
@click.option(
    "--ref", type=str, multiple=True, help="references for source. e.g. --ref user:llama --ref location:ocean"
)
@click.option("--timestamp", type=str, help=TIMESTAMP_DESCRIPTION)
@click.option("--security", required=True, type=str, help=SECURITY_STRING_DESCRIPTION)
@click.option("--extract", is_flag=True, show_default=True, default=False, help="submitted files are trusted archives")
@click.option("--extract-password", type=str, help="password for trusted archive to be extracted with")
def put(
    y: bool,
    path: str,
    timestamp: str,
    security: str,
    source: str,
    ref: list[str],
    extract: bool,
    extract_password: str,
):
    """Upload all files in PATH to Azul SOURCE."""
    split_refs = [x.split(":", 1) for x in ref]
    refs = {x[0]: x[1] for x in split_refs}
    _shared_submit(
        y,
        path,
        security=security,
        timestamp=timestamp,
        source=source,
        source_refs=refs,
        extract=extract,
        extract_password=extract_password,
    )


@binaries.command()
@click.argument("filename")
@click.argument("source")
@click.option("-y", is_flag=True, show_default=True, default=False, help="automatic confirmation")
@click.option(
    "--ref", type=str, multiple=True, help="references for source. e.g. --ref user:llama --ref location:ocean"
)
@click.option("--timestamp", type=str, help=TIMESTAMP_DESCRIPTION)
@click.option("--security", required=True, type=str, help=SECURITY_STRING_DESCRIPTION)
def put_stdin(y: bool, filename: str, source: str, ref: list[str], timestamp: str, security: str):
    """Upload a file from stdin into an Azul source.

    FILENAME is the name of the file in Azul, and SOURCE is the ID of the source to upload the file to.
    """
    split_refs = [x.split(":", 1) for x in ref]
    refs = {x[0]: x[1] for x in split_refs}
    security = security if security else ""

    if not timestamp:
        timestamp = pendulum.now(pendulum.UTC).to_iso8601_string()
    else:
        parsed_timestamp = pendulum.parse(timestamp)
        if isinstance(parsed_timestamp, DateTime):
            timestamp = parsed_timestamp.to_iso8601_string()
        else:
            click.echo(f"Provided timestamp {timestamp} is in an invalid format, it must be a DateTime.")
            raise Exception("Provided timestamp")

    click.echo(f"Filename: {filename}")
    click.echo(f"Source: {source}")
    click.echo(f"References: {refs}")
    click.echo(f"Timestamp: {timestamp}")
    click.echo(f"Security: {security}")

    if not y and not click.confirm("Proceed with upload of file?"):
        sys.exit(1)

    # Read file in chunks into a spooled temporary file.
    chunk_size = 1024 * 1024 * 1024
    with SpooledTemporaryFile(max_size=chunk_size) as spooledFile:
        while chunk := sys.stdin.buffer.read(chunk_size):
            spooledFile.write(chunk)
        spooledFile.seek(0)

        resp = api.binaries_data.upload(
            spooledFile,
            security=security,
            source_id=source,
            filename=filename,
            timestamp=timestamp,
            references=refs,
        )
        click.echo(f"{filename} - {resp.sha256}")


@binaries.command(help="""Get a binary's metadata from Azul by SHA256.""")
@click.argument("sha256")
@click.option(
    "-o",
    "--output",
    help="Output to a file - use '-' for stdout.",
    default="-",
    show_default=True,
    type=click.Path(file_okay=True, dir_okay=False, allow_dash=True),
)
@click.option(
    "--pretty/--no-pretty",
    help="Render stdout output coloured (default true if terminal, else false).",
    default=os.isatty(sys.stdout.fileno()),
)
def get_meta(sha256: str, output: str, pretty: bool):
    """Get metadata for a binary."""
    entity = api.binaries_meta.get_meta(sha256)

    if output == "-":
        _print_model(entity, pretty)
    else:
        click.echo(f"saving output to path {output}", err=True)
        with open(output, "w") as f:
            f.write(entity.model_dump_json(indent=4))


@binaries.command(
    help="""
Find and download samples from Azul.
Combining multiple filters may lead to unexpected results.
You can only query multiple attributes over a single authors document.
"""
)
@click.option("-o", "--output", help="output folder")
@click.option("--term", help="search term (refer to UI Explore for suggested search terms)", default="")
@click.option("--max", help="max number of entities to retrieve", default=100)
@click.option(
    "--sort-by",
    default=None,
    type=click.Choice(
        [
            str(models_restapi.FindBinariesSortEnum.score),
            str(models_restapi.FindBinariesSortEnum.source_timestamp),
            str(models_restapi.FindBinariesSortEnum.timestamp),
        ]
    ),
    help="What property to use when sorting results",
)
@click.option(
    "--sort-asc", default=False, is_flag=True, show_default=True, help="sort by ascending rather than descending."
)
def get(output: str, term: str, max: int, sort_by: models_restapi.FindBinariesSortEnum, sort_asc: bool):
    """Get all samples matching the criteria and optionally download the files to an output folder."""
    if output:
        click.echo(f"saving output to folder {output}")
    else:
        click.echo("no output folder provided, skip download")

    params = {"term": term, "max_entities": max, "sort_prop": sort_by, "sort_asc": sort_asc}
    kwargs = {x: y for (x, y) in params.items() if y is not None}
    entity = api.binaries_meta.find(**kwargs)

    # create output folder
    if output:
        click.echo(f"download to folder {output}")
        if not os.path.exists(output):
            click.echo(f"creating directory: {output}")
            os.mkdir(output)
        if not os.path.isdir(output):
            raise Exception(f"supplied path is not a directory: {output}")

    # print and save found binary
    for hit in entity.items:
        click.echo(hit.sha256)
        if output:
            content = api.binaries_data.download(hit.sha256)
            if content:
                with open(os.path.join(output, f"{hit.sha256}.cart"), "wb") as f:
                    f.write(content)
            else:
                click.echo("content not found")


@binaries.command(
    help="""
Downloaded the provided sha256's from azul in Cart form.
All provided arguments will be considered sha256's.
You can also optionally provide and output folder to not download to the current directory.
"""
)
@click.argument("sha256s", nargs=-1, type=str)
@click.option("-o", "--output", help="output folder")
def download(sha256s: tuple[str], output: str):
    """Download the provided sha256's from Azul."""
    if output:
        click.echo(f"saving output to folder {output}")
    else:
        click.echo("saving file to current directory")
        output = ""

    invalid_sha256s = []
    valid_sha256s = []
    for sha in sha256s:
        if not re.search(api.binaries_data.SHA256_regex, sha):
            invalid_sha256s.append(sha)
        else:
            valid_sha256s.append(sha)

    if len(invalid_sha256s) > 0:
        click.echo(f"\nIgnoring the following provided invalid sha256s:\n  {'\n  '.join(invalid_sha256s)}\n")

    if not valid_sha256s:
        click.echo("No valid sha256's provided stopping now.")
        return

    click.echo(f"Downloading the following sha256's:\n  {'\n  '.join(valid_sha256s)}\n")
    successful_download_paths = []
    for sha in valid_sha256s:
        # Download file and account for fail cases.
        try:
            raw_bytes = api.binaries_data.download(sha)
        except BadResponse404:
            click.echo(f"Error: Could not download the sha256 {sha} because it does not exist in Azul.")
            continue
        except BadResponse as e:
            click.echo(f"Error: Could not download the sha256 {sha} with error {e}")
            continue
        except Exception as e:
            click.echo(f"Error: Could not download the sha256 {sha} with unexpected error {e}")
            continue

        try:
            new_path = os.path.join(output, sha + ".cart")
            with open(new_path, "wb") as f:
                f.write(raw_bytes)

            successful_download_paths.append(new_path)
        except Exception as e:
            click.echo(f"Could not save the sha256 {sha} after successful download with error\n {e}")

    # Nicer new line formatting.
    if len(successful_download_paths) != len(valid_sha256s):
        click.echo("")

    click.echo(f"Download completed for the following files:\n  {'\n  '.join(successful_download_paths)}")


cli.add_command(_client_config)
cli.add_command(binaries)
cli.add_command(binaries, name="b")
cli.add_command(security)
cli.add_command(sources)
cli.add_command(plugins)
if __name__ == "__main__":
    cli()



import time

from azul_bedrock import models_restapi

from . import module_sha256s, module_source
from .base_test import BaseApiTest


class TestPurgeApi(BaseApiTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.calculated_username = cls.api.users.get_user_info().username

    def test_purge_submission(self):
        """Test you could purge a submission.

        NOTE - you must be admin for this to work.
        """
        # Find needed metadata for submission
        binary_meta = self.api.binaries_meta.get_meta(module_sha256s[0])
        submission_meta = binary_meta.sources[0].direct[0]

        # Simulate purge But timestamp isn't provided
        self.assertRaises(
            ValueError, self.api.purge.purge_submission, submission_meta.track_source_references, timestamp=None
        )

        # Simulate purge But timestamp is invalid
        self.assertRaises(
            ValueError, self.api.purge.purge_submission, submission_meta.track_source_references, timestamp="None"
        )
        # Bad date that should get 0 results.
        resp_bad_date = self.api.purge.purge_submission(submission_meta.track_source_references, timestamp="2024")
        self.assertEqual(resp_bad_date.events, 0)

        # Simulate purge for timestamp
        resp_time = self.api.purge.purge_submission(
            submission_meta.track_source_references, timestamp=submission_meta.timestamp
        )
        self.assertGreaterEqual(resp_time.events, 1)

    def test_purge_link(self):
        """Simulate purging a child binary because actually purging it would cause race condition issues.

        NOTE - you must be admin for this to work.
        """
        # Add a child
        parent = module_sha256s[0]
        child_entity = self.api.binaries_data.upload_child(
            b"child_file_for_file1",
            parent,
            {"extracted": "by-test-user"},
            filename="child1.txt",
            security=self.default_security,
        )

        child_id = child_entity.sha256
        self.assertEqual(child_id, "30c275c36b2960732a63cf800408cde629c6b8b136b4f01d99610609696689ad")
        time.sleep(2)

        binary_meta = self.api.binaries_meta.get_meta(child_entity.sha256)
        self.assertGreaterEqual(len(binary_meta.parents), 1)

        # Simulate purging the child.
        resp: models_restapi.PurgeSimulation = self.api.purge.purge_link(binary_meta.parents[0].track_link)
        self.assertGreaterEqual(resp.events, 1)



"""Api wrapper."""

import logging
import sys

from azul_client import config, oidc

from . import (
    binaries_data,
    binaries_meta,
    features,
    plugins,
    purge,
    security,
    sources,
    statistics,
    users,
)

logging.basicConfig(stream=sys.stdout, level=logging.WARNING)


class Api:
    """Contains api implementation instances.

    This is so we can chain different API calls without having to
    juggle multiple objects.
    """

    def __init__(self, conf: config.Config | None = None) -> None:
        # api support
        self.config = conf if conf else config.get_config()
        self.auth = oidc.OIDC(self.config)

        self._api_implementations = []
        # api implementations
        self.binaries_data = binaries_data.BinariesData(self.config, self.auth.get_client)
        self._api_implementations.append(self.binaries_data)

        self.binaries_meta = binaries_meta.BinariesMeta(self.config, self.auth.get_client)
        self._api_implementations.append(self.binaries_meta)

        self.features = features.Features(self.config, self.auth.get_client)
        self._api_implementations.append(self.features)

        self.plugins = plugins.Plugins(self.config, self.auth.get_client)
        self._api_implementations.append(self.plugins)

        self.purge = purge.Purge(self.config, self.auth.get_client)
        self._api_implementations.append(self.purge)

        self.security = security.Security(self.config, self.auth.get_client)
        self._api_implementations.append(self.security)

        self.sources = sources.Sources(self.config, self.auth.get_client)
        self._api_implementations.append(self.sources)

        self.statistics = statistics.Statistics(self.config, self.auth.get_client)
        self._api_implementations.append(self.statistics)

        self.users = users.Users(self.config, self.auth.get_client)
        self._api_implementations.append(self.users)

    def get_excluded_security(self) -> list[str]:
        """Get excluded security."""
        return self._excluded_security

    def set_excluded_security(self, security_list: list[str]):
        """Set excluded security's to the provided values."""
        self._excluded_security = security_list

        # Set excluded security for all API's
        for api in self._api_implementations:
            api._excluded_security = self._excluded_security



import os
from unittest import TestCase

import httpx
import respx
from azul_bedrock import models_restapi
from azul_bedrock import models_settings as azs
from click.testing import CliRunner

from azul_client import Api, client
from azul_client import config as mconfig


class ClientTestOther(TestCase):
    def setUp(self):
        mconfig.location = ""
        cur_api = Api(
            mconfig.Config(
                auth_type="none",
                oidc_url="http://localhost",
                auth_client_id="servicer",
            )
        )
        client.api = cur_api
        self.api = cur_api

        self.cli = CliRunner()

        return super().setUp()

    @respx.mock
    def test_security(self):
        """Test listing out security works as expected"""
        security_resp = {
            "extra": {"extra extra": "read all about it"},
            "presets": [
                "OFFICIAL",
                "OFFICIAL TLP:CLEAR",
                "OFFICIAL TLP:GREEN",
                "MORE OFFICIAL FIREBALL REL:APPLE",
            ],
        }
        # Mock security response
        respx.get(f"{self.api.config.azul_url}/api/v0/security").mock(
            httpx.Response(status_code=200, json=security_resp)
        )

        # Standard security listing
        res = self.cli.invoke(client.security)
        self.assertIn("OFFICIAL", res.stdout)
        self.assertIn("OFFICIAL TLP:CLEAR", res.stdout)
        self.assertIn("OFFICIAL TLP:GREEN", res.stdout)
        self.assertIn("MORE OFFICIAL FIREBALL REL:APPLE", res.stdout)
        self.assertNotIn("extra", res.stdout)

        # Try with full option
        res = self.cli.invoke(client.security, ["--full"])
        self.assertIn("OFFICIAL", res.stdout)
        self.assertIn("OFFICIAL TLP:CLEAR", res.stdout)
        self.assertIn("OFFICIAL TLP:GREEN", res.stdout)
        self.assertIn("MORE OFFICIAL FIREBALL REL:APPLE", res.stdout)
        self.assertIn("extra", res.stdout)

    @respx.mock
    def test_list_sources(self):
        """Test that sources list out as expected."""
        source_resp = {
            "testing": azs.Source(
                references=[
                    azs.Source.SourceReference(
                        name="custom_ref_1",
                        required=True,
                        description="testing description 1",
                    )
                ]
            ).model_dump(),
            "other": azs.Source(
                references=[
                    azs.Source.SourceReference(
                        name="custom_ref_2",
                        required=False,
                        description="other reference field",
                    )
                ]
            ).model_dump(),
        }
        respx.get(f"{self.api.config.azul_url}/api/v0/sources").mock(
            httpx.Response(status_code=200, json={"data": source_resp})
        )

        # Listing just the source name works as expected.
        res = self.cli.invoke(client.sources_list)
        # For debugging issues.
        print(res)

        self.assertIn("testing", res.stdout)
        self.assertIn("other", res.stdout)
        self.assertNotIn("custom_ref_1", res.stdout)
        self.assertNotIn("custom_ref_2", res.stdout)

        # List will all the information about all sources.
        res = self.cli.invoke(client.sources_full)
        print(res)
        self.assertIn("testing", res.stdout)
        self.assertIn("other", res.stdout)
        self.assertIn("custom_ref_1", res.stdout)
        self.assertIn("custom_ref_2", res.stdout)

        # Test case of listing a specific source
        res = self.cli.invoke(client.sources_info, ["testing"])
        print(res)
        self.assertIn("testing", res.stdout)
        self.assertIn("custom_ref_1", res.stdout)
        self.assertIn("testing description 1", res.stdout)

        self.assertNotIn("other", res.stdout)
        self.assertNotIn("custom_ref_2", res.stdout)

    @respx.mock
    def test_list_plugins(self):
        mock_resp: list[models_restapi.LatestPluginWithVersions] = [
            models_restapi.LatestPluginWithVersions(
                versions=["2023", "2024", "2025"],
                newest_version=models_restapi.PluginEntity(
                    name="dummy",
                    version="2025",
                    description="plugin description",
                    security="OFFICIAL",
                    category="plugin",
                ),
            ).model_dump(),
        ]
        respx.get(f"{self.api.config.azul_url}/api/v0/plugins").mock(
            httpx.Response(status_code=200, json={"data": mock_resp})
        )
        mock_single_resp = models_restapi.PluginInfo(
            num_entities=10,
            plugin=models_restapi.PluginEntity(
                name="dummy",
                version="2023",
                description="different desc so I can verify",
                security="OFFICIAL",
                category="plugin",
            ),
            status=[],
        ).model_dump()
        respx.get(f"{self.api.config.azul_url}/api/v0/plugins/dummy/versions/2023").mock(
            httpx.Response(status_code=200, json={"data": mock_single_resp})
        )

        # List all plugins
        res = self.cli.invoke(client.plugins_list)
        print(res.stdout)
        self.assertIn("dummy 2025", res.stdout)

        # Get detail about latest version of this plugin.
        res = self.cli.invoke(client.plugin_info, ["dummy"])
        self.assertIn("dummy", res.stdout)
        self.assertIn("2025", res.stdout)
        self.assertIn("plugin description", res.stdout)
        self.assertIn("OFFICIAL", res.stdout)

        # Get detail about older version of this plugin.
        res = self.cli.invoke(client.plugin_info, ["dummy", "--version", "2023"])
        print(res)
        self.assertIn("dummy", res.stdout)
        self.assertIn("2023", res.stdout)
        self.assertIn("different desc so I can verify", res.stdout)
        self.assertIn("OFFICIAL", res.stdout)



"""Base API class used by all other Azul API classes."""

import json
import logging
from http import HTTPMethod
from typing import Any, Callable, TypeVar

import httpx
from azul_bedrock import models_restapi
from pydantic import BaseModel, TypeAdapter

from azul_client import config, exceptions

T = TypeVar("T", bound=BaseModel)


class BaseApiHandler:
    """Base class for handling all restapi calls with generic functionality."""

    def __init__(self, cfg: config.Config, get_client: Callable[[], httpx.Client]):
        self.logger = logging.getLogger(type(self).__name__)
        self.cfg = cfg
        self.__get_client = get_client
        self._last_meta = None
        self.__excluded_security = None

    @property
    def _excluded_security(self) -> list[str]:
        """Get excluded security (recommended to get this at the API level)."""
        return self.__excluded_security

    @_excluded_security.setter
    def _excluded_security(self, security_list: list[str]):
        """Set excluded security's to the provided values (internal use only get directly get from API level)."""
        self.__excluded_security = security_list

    def get_meta_from_last_request(self) -> models_restapi.Meta | None:
        """Get the metadata from the last request that was made, if there is no metadata return None."""
        if self._last_meta:
            return models_restapi.Meta.model_validate(self._last_meta)
        return None

    def __request_to_client(
        self,
        method: HTTPMethod,
        url: str,
        params: httpx.QueryParams | dict | None = None,
        json: Any = None,
        timeout: int | None = None,
    ) -> httpx.Response:
        if timeout is None:
            timeout = httpx.USE_CLIENT_DEFAULT
        if self._excluded_security:
            if not params:
                params = dict()
            params["x"] = self._excluded_security
        if method == HTTPMethod.GET:
            if json is not None:
                raise ValueError("Get request cannot accept a body parameter.")
            return self.__get_client().get(url, params=params, timeout=timeout)
        elif method == HTTPMethod.POST:
            return self.__get_client().post(url, params=params, json=json, timeout=timeout)
        elif method == HTTPMethod.DELETE:
            if json is not None:
                raise ValueError("Delete request cannot accept a body parameter.")
            return self.__get_client().delete(url, params=params, timeout=timeout)
        raise ValueError(
            f"The provided method '{method}' is invalid it must be one of "
            + f"{', '.join([HTTPMethod.POST, HTTPMethod.POST])}"
        )

    def _request_with_pydantic_model_response(
        self,
        *,
        method: HTTPMethod,
        url: str,
        response_model: type[T],
        params: httpx.QueryParams | dict | None = None,
        json: Any = None,
        get_data_only: bool = False,
        timeout: int | None = None,
    ) -> T:
        """Generic Handler for requests.

        :param HTTPMethod method: HTTP method to use only GET and POST are supported.
        :param str url: url to post or get to.
        :param BaseModel response_model: Expected Pydantic response model.
        :param httpx.QueryParams params: parameters for the request.
        :param Any json: raw body that will be json encoded and sent with POST request, doesn't work with get requests.
        :param bool get_data_only: will extract out the 'data' key from the response and throws an
        exception if there is no data field.
        """
        # API requests should always clear last requests metadata
        self._last_meta = None
        resp = self.__request_to_client(method=method, url=url, params=params, json=json, timeout=timeout)

        if resp.status_code != 200 and resp.status_code != 206:
            raise exceptions.bad_response(resp)

        raw_content: str | bytes = resp.content
        if get_data_only:
            raw_content = self._get_response_data(resp)

        try:
            if isinstance(response_model, TypeAdapter):
                return response_model.validate_json(raw_content)
            return response_model.model_validate_json(raw_content)
        except Exception:
            self.logger.error(f"Failed to deserialize pydantic model {type(response_model)}.")
            self.logger.error(f"Response started with {raw_content[:500]}")
            raise

    def _request(
        self,
        *,
        method: HTTPMethod,
        url: str,
        params: httpx.QueryParams = None,
        json: Any = None,
        timeout: int | None = None,
    ) -> httpx.Response:
        """Send a http request with the provided method to azul, and provide the json response."""
        # API requests should always clear last requests metadata
        self._last_meta = None
        resp = self.__request_to_client(method=method, url=url, params=params, json=json, timeout=timeout)

        if resp.status_code != 200 and resp.status_code != 206:
            raise exceptions.bad_response(resp)
        return resp

    def _request_upload(self, *, url: str, params: dict, files: dict, data: dict, timeout: int) -> httpx.Response:
        """Special request type for uploading files."""
        # API requests should always clear last requests metadata
        self._last_meta = None
        return self.__get_client().post(url=url, params=params, files=files, data=data, timeout=timeout)

    def _generic_head_request(self, url: str) -> bool:
        """Generic request to check if a resource exists."""
        # API requests should always clear last requests metadata
        self._last_meta = None
        resp = self.__get_client().head(url)
        if resp.status_code == 404:
            return False
        if resp.status_code != 200 and resp.status_code != 206:
            raise exceptions.bad_response(resp)
        return True

    def _get_response_data(self, resp: httpx.Response) -> str:
        """Get the 'data' key from a response when it may also have a 'metadata' field."""
        json_response = resp.json()
        data = json_response.get("data", None)
        self._last_meta = json_response.get("meta", None)
        if data is None:
            raise Exception("Response has no 'data' key and 'data cannot be extracted.")

        return json.dumps(data)

    def filter_none_values(self, params: dict) -> dict:
        """Takes a dictionary and filters out all keys with None values."""
        for k, v in list(params.items()):
            if v is None:
                params.pop(k)
        return params



import json

import pytest
from pytest_httpx import HTTPXMock

from azul_client import Api, Config
from azul_client import config as mconfig


@pytest.fixture
def api() -> Api:
    mconfig.location = ""
    return Api(
        Config(
            auth_type="none",
            azul_url="http://localhost:8123",
        )
    )


def test_normalise_basic(api: Api, httpx_mock: HTTPXMock):
    httpx_mock.add_response(
        method="POST",
        url=f"{api.config.azul_url}/api/v1/security/normalise",
        status_code=200,
        content=b'"OFFICIAL TLP:GREEN"',
    )
    normalised_security = api.security.normalise("OFFICIAL//TLP:GREEN")
    body = json.loads(httpx_mock.get_request().read())
    assert body == {"security": "OFFICIAL//TLP:GREEN"}
    print(normalised_security)
    assert normalised_security == "OFFICIAL TLP:GREEN"



import io
import time

from azul_bedrock import models_network as azm
from azul_bedrock import models_restapi

from azul_client.api.binaries_data import AugmentedStream
from azul_client.exceptions import BadResponse404

from .base_test import BaseApiTest


class TestBinaryDataOther(BaseApiTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        # NOTE - not recommend to change this by even a character as it will break a lot of tests
        cls.std_content = io.BytesIO(b"Content for strings\n and hex view\n that is testable!")
        cls.std_kwargs = {
            "source_id": cls.known_source,
            "filename": "test_binary_other_file.txt",
            "security": cls.default_security,
            "references": cls.upload_ref,
        }
        cls.aug_value = b"augvalue"

        # Create a binary with an augmented stream.
        resp = cls.api.binaries_data.upload(
            cls.std_content,
            **cls.std_kwargs,
            augmented_streams=[
                AugmentedStream(
                    label=azm.DataLabel.TEST,
                    file_name="aug_stream1.txt",
                    contents_file_path=io.BytesIO(cls.aug_value),
                ),
            ],
            refresh=True,
        )

        # Store sha256's in text because augmented stream hash isn't in response.
        cls.binary_other_sha256 = "9b9093d00cd0cb1b2542642c5042badccaa49b8bf258731141bba5633c0c6bd5"
        cls.binary_aug_sha256 = "436f659951e835408044d665ec174e6bf8d4ba1899c556e41a48c54b24883717"

        found = False
        max_total_attempts = 3
        while not found:
            max_total_attempts += 1
            start_time = time.time()
            max_wait = 15  # wait up to 15 seconds for binaries to be in the system.
            while start_time + max_wait > time.time():
                if cls.api.binaries_meta.check_meta(cls.binary_other_sha256):
                    if cls.api.binaries_data.check_data(cls.binary_other_sha256):
                        found = True
                        break
                time.sleep(2)

        if not found:
            raise Exception(
                f"Failed to verify binary_data_other was uploaded to azul in {max_wait} seconds.\n !"
                + f"File hashe is {cls.binary_other_sha256}"
            )

    def test_verify_sha256s_length(self):
        """Ensure the sha256's are the appropriate length.

        This covers the tests that loop over sha256 to make sure their assertions are actually getting called.
        For the actual length refer to the module setup in __init__
        """
        self.assertGreaterEqual(len(self.sha256s), 10)

    def test_bulk_download(self):
        content = self.api.binaries_data.download_bulk(self.sha256s)
        for sha in self.sha256s:
            self.assertIn(bytes(sha, "utf-8"), content)

    def test_download(self):
        response1 = self.api.binaries_data.download(self.sha256s[0])
        # Ensure first 4 bytes are the start of the cart file.
        self.assertEqual(response1[:4], b"CART")
        response2 = self.api.binaries_data.download(self.sha256s[1])
        self.assertEqual(response2[:4], b"CART")

    def test_check_data(self):
        for i, sha in enumerate(self.sha256s):
            exists = self.api.binaries_data.check_data(sha)
            self.assertEqual(exists, True, f"The sha256 {sha} at index {i} does not exist!")

        exists = self.api.binaries_data.check_data(self.non_existent_sha256)
        self.assertEqual(exists, False)

    def test_expedite_processing(self):
        """Expedite processing."""
        first_sha256 = self.sha256s[0]
        self.api.binaries_data.expedite_processing(first_sha256)
        self.api.binaries_data.expedite_processing(first_sha256, bypass_cache=True)
        self.api.binaries_data.expedite_processing(self.non_existent_sha256)

    def test_download_augmented_stream(self):
        aug_data = self.api.binaries_data.download_augmented_stream(self.binary_other_sha256, self.binary_aug_sha256)
        self.assertEqual(aug_data, self.aug_value)

        # Non-existent binary and augmented stream
        with self.assertRaises(BadResponse404):
            aug_data = self.api.binaries_data.download_augmented_stream(
                self.non_existent_sha256, self.non_existent_sha256
            )

    def test_download_hex(self):
        # Test getting the whole hex file.
        resp = self.api.binaries_data.download_hex(self.binary_other_sha256)
        self.assertEqual(len(resp.hex_strings), 4)
        self.assertEqual(resp.hex_strings[0].address, 0)
        hex_val = ["43", "6F", "6E", "74", "65", "6E", "74", "20", "66", "6F", "72", "20", "73", "74", "72", "69"]
        self.assertEqual(resp.hex_strings[0].hex, hex_val)

        self.assertEqual(resp.header.address, "ADDRESS")

        header_val = ["00", "01", "02", "03", "04", "05", "06", "07", "08", "09", "0A", "0B", "0C", "0D", "0E", "0F"]
        self.assertEqual(resp.header.hex, header_val)

        self.assertEqual(resp.next_offset, 52)
        self.assertEqual(resp.content_length, 52)
        self.assertFalse(resp.has_more)

        # Test shortform
        resp = self.api.binaries_data.download_hex(self.binary_other_sha256, shortform=True)
        self.assertEqual(len(resp.hex_strings), 4)
        self.assertEqual(resp.hex_strings[0].address, 0)
        self.assertEqual(resp.hex_strings[0].hex, "436f 6e74 656e 7420 666f 7220 7374 7269")

        self.assertEqual(resp.header.address, "ADDRESS")
        self.assertEqual(resp.header.hex, "0001 0203 0405 0607 0809 0a0b 0c0d 0e0f")

        self.assertEqual(resp.next_offset, 52)
        self.assertEqual(resp.content_length, 52)
        self.assertFalse(resp.has_more)

        # Test all other parameters
        resp = self.api.binaries_data.download_hex(self.binary_other_sha256, offset=1, max_bytes_to_read=10)
        print(resp)
        self.assertEqual(len(resp.hex_strings), 1)
        self.assertEqual(resp.hex_strings[0].hex, hex_val[1:11] + ["", "", "", "", "", ""])

        self.assertTrue(resp.next_offset, 11)
        self.assertTrue(resp.content_length, 52)
        self.assertTrue(resp.has_more)

    def assert_binary_strings_equal(
        self, response: models_restapi.BinaryStrings, expected: models_restapi.BinaryStrings, test_ref: str
    ):
        print("---")
        print(test_ref + ":")
        print(response)
        self.assertEqual(response.has_more, expected.has_more, "has_more doesn't match")
        self.assertEqual(response.next_offset, expected.next_offset, "Offsets don't match")
        self.assertEqual(response.strings, expected.strings, "strings don't match")
        print("---")

    def test_get_strings(self):
        og_expected_result = models_restapi.BinaryStrings(
            strings=[
                models_restapi.SearchResult(
                    string="Content for strings", offset=0, length=19, encoding=models_restapi.SearchResultType.ASCII
                ),
                models_restapi.SearchResult(
                    string=" and hex view", offset=20, length=13, encoding=models_restapi.SearchResultType.ASCII
                ),
                models_restapi.SearchResult(
                    string=" that is testable!", offset=34, length=18, encoding=models_restapi.SearchResultType.ASCII
                ),
            ],
            has_more=False,
            next_offset=52,
        )
        resp = self.api.binaries_data.get_strings(self.binary_other_sha256)
        self.assert_binary_strings_equal(resp, og_expected_result, "base case")

        # All the different possible parameters
        # Min_length
        expected_result = og_expected_result.model_copy(deep=True)
        expected_result.strings = [expected_result.strings[0], expected_result.strings[2]]
        self.assert_binary_strings_equal(
            self.api.binaries_data.get_strings(self.binary_other_sha256, min_length=15), expected_result, "min length"
        )

        # Max Length
        expected_result = og_expected_result.model_copy(deep=True)
        expected_result.strings = [expected_result.strings[1]]
        self.assert_binary_strings_equal(
            self.api.binaries_data.get_strings(self.binary_other_sha256, max_length=15), expected_result, "max length"
        )

        # Offset
        expected_result = og_expected_result.model_copy(deep=True)
        expected_result.strings = expected_result.strings[1:]
        self.assert_binary_strings_equal(
            self.api.binaries_data.get_strings(self.binary_other_sha256, offset=19), expected_result, "Offset"
        )

        # Max bytes to read
        expected_result = models_restapi.BinaryStrings(
            strings=[
                models_restapi.SearchResult(
                    string="Conte", offset=0, length=5, encoding=models_restapi.SearchResultType.ASCII
                )
            ],
            has_more=False,
            next_offset=5,
        )
        self.assert_binary_strings_equal(
            self.api.binaries_data.get_strings(self.binary_other_sha256, max_bytes_to_read=5),
            expected_result,
            "Max bytes to read",
        )

        # take_n_strings
        expected_result = models_restapi.BinaryStrings(
            strings=[
                models_restapi.SearchResult(
                    string="Content for strings", offset=0, length=19, encoding=models_restapi.SearchResultType.ASCII
                )
            ],
            has_more=True,
            next_offset=20,
        )
        self.assert_binary_strings_equal(
            self.api.binaries_data.get_strings(self.binary_other_sha256, take_n_strings=1),
            expected_result,
            "take_n_strings",
        )

        # filter first string
        expected_result = og_expected_result.model_copy(deep=True)
        expected_result.strings = [expected_result.strings[0]]
        self.assert_binary_strings_equal(
            self.api.binaries_data.get_strings(self.binary_other_sha256, filter="Content for strings"),
            expected_result,
            "filter first string",
        )

        # filter regex
        expected_result = og_expected_result.model_copy(deep=True)
        expected_result.strings = [expected_result.strings[2]]
        self.assert_binary_strings_equal(
            self.api.binaries_data.get_strings(self.binary_other_sha256, regex=r"[^a-zA-Z\d\s:]"),
            expected_result,
            "filter regex",
        )

    def test_search_hex(self):
        og_expected_result = models_restapi.BinaryStrings(
            strings=[
                models_restapi.SearchResult(
                    string="te", offset=3, length=2, encoding=models_restapi.SearchResultType.Hex
                ),
                models_restapi.SearchResult(
                    string="te", offset=43, length=2, encoding=models_restapi.SearchResultType.Hex
                ),
            ],
            has_more=False,
            next_offset=52,
        )
        search_string = "7465"
        # Searching for "te" in hex
        resp = self.api.binaries_data.search_hex(self.binary_other_sha256, search_string)
        self.assert_binary_strings_equal(resp, og_expected_result, "base case")

        # offset
        expected_result = og_expected_result.model_copy(deep=True)
        expected_result.strings = [expected_result.strings[1]]
        self.assert_binary_strings_equal(
            self.api.binaries_data.search_hex(self.binary_other_sha256, search_string, offset=8),
            expected_result,
            "offset",
        )

        # max_bytes_to_read
        expected_result = og_expected_result.model_copy(deep=True)
        expected_result.strings = [expected_result.strings[0]]
        expected_result.next_offset = 6
        self.assert_binary_strings_equal(
            self.api.binaries_data.search_hex(self.binary_other_sha256, search_string, max_bytes_to_read=6),
            expected_result,
            "max_bytes_to_read",
        )

        # take_n_hits
        expected_result = og_expected_result.model_copy(deep=True)
        expected_result.strings = [expected_result.strings[0]]
        self.assert_binary_strings_equal(
            self.api.binaries_data.search_hex(self.binary_other_sha256, search_string, take_n_hits=1),
            expected_result,
            "take_n_hits",
        )
